# -*- coding: utf-8 -*-
#
# Tencent is pleased to support the open source community by making QTA available.
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the BSD 3-Clause License (the "License"); you may not use this 
# file except in compliance with the License. You may obtain a copy of the License at
# 
# https://opensource.org/licenses/BSD-3-Clause
# 
# Unless required by applicable law or agreed to in writing, software distributed 
# under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#
"""
资源管理模块

包括：协作机管理及测试资源文件管理
"""

import os
import sys
import locale
import uuid
import pkg_resources
import csv

from testbase import context
import testbase.logger as logger
from testbase.conf import settings
from testbase.util import smart_text, codecs_open
from testbase.testresult import EnumLogLevel

os_encoding = locale.getdefaultlocale()[1]
if not os_encoding:
    os_encoding = 'utf8'

RESMGR_BACKEND_ENTRY_POINT = "qtaf.resmgr_backend"
resmgr_backend_types = {}

class ResourceNotAvailable(Exception):
    """没有可用的资源
    """
    pass

class DownloadFileError(Exception):
    """download file failed
    """
    def __init__(self, url, status_code, msg, headers, data):
        self.url = url
        self.status_code = status_code
        self.msg = msg
        self.headers = headers
        self.data = data
        
    def __str__(self):
        return "downloading file failed for: %s %s %s\nheaders=%s\nbody=%s" % (self.url,
                                                             self.status_code,
                                                             self.msg,
                                                             self.headers,
                                                             smart_text(self.data))



class Session(object):
    """会话
    """
    def __init__(self, backend, session_id):
        """构造函数
        """
        self._backend = backend
        self._id = session_id
        
    def acquire_resource(self, res_type, res_group=None, condition=None):
        """申请资源

        :param res_type: 资源类型
        :type res_type: str
        :param res_group: 资源分组
        :type res_type: str
        :param condition: 资源属性匹配
        :type condition: dict
        :returns: 资源
        :rtypes: dict
        """
        tc = context.current_testresult()
        try:
            resource = self._backend.acquire_resource(self._id, res_type, res_group, condition)
        except ResourceNotAvailable:
            if tc:
                tc.log_record(EnumLogLevel.RESNOTREADY, "acquire resource (res_type:%s, res_group:%s, condition:%s) failed" % (res_type, res_group, condition))
            raise
        else:
            if tc:
                extra = {}
                extra["res_type"] = res_type
                if not isinstance(resource, dict):
                    raise ValueError("Resource record should be a dictionary")
                if "id" not in resource:
                    raise ValueError("Resource record should be a dictionary with key 'id'")
                extra["resource_id"] = resource["id"]
                tc.log_record(EnumLogLevel.RESOURCE, "acquire resource (res_type:%s, res_group:%s, condition:%s) successfully" % (res_type, res_group, condition), extra)
            return resource

    def release_resource(self, res_type, resource_id):
        """释放资源

        :param res_type: 资源类型
        :type res_type: str
        :param resource_id: 资源ID
        :type resource_id: str
        """
        return self._backend.release_resource(self._id, res_type, resource_id)
        
    def destroy(self):
        """销毁该会话（全部占用的资源会释放）
        """
        if not self._id:
            return
        res = self._backend.destroy_session(self._id)
        self._id = None
        return res
        
    def get_file(self, path):
        """获取测试文件资源

        :param path: 文件引用路径（相对路径）
        :returns: 文件路径
        :rtypes: str
        """
        return self._backend.get_file(path)
    
    def list_dir(self,path):
        """获取目录下文件列表

        :param path: 目录引用路径（相对路径）
        :returns: 一个包含该目录下所有文件的绝对路径的list 
        :rtypes: list[str]
        """
        return self._backend.list_dir(path)
    
    def walk(self,path):
        """获取目录下文件列表

        :param path: 相对于resources目录的路径，用于遍历文件夹
        :type  path: str
        :return iterators: iterator of (dir_path, dirnames, filenames) tuples
        :type   iterators: iterator
        """
        return self._backend.walk(path)

    
class IResourceManagerBackend(object):
    """测试资源管理后端接口定义
    """
    def create_session(self, testcase=None):
        """创建会话

        :param testcase: 使用的测试用例
        :type testcase: TestCase
        :returns: 会话ID
        :rtypes: str
        """
        raise NotImplementedError()
        
    def destroy_session(self, session_id):
        """销毁会话

        :param session_id: 会话ID
        :type session_id: str
        """
        raise NotImplementedError()
    
    def acquire_resource(self, session_id, res_type, res_group, condition ):
        """申请资源

        :param session_id: 会话ID
        :type session_id: str
        :param res_type: 资源类型
        :type res_type: str
        :param res_group: 资源分组
        :type res_type: str
        :param condition: 资源属性匹配
        :type condition: dict
        :returns: 资源
        :rtypes: dict
        """
        raise NotImplementedError()
        
    def release_resource(self, session_id, res_type, resource_id):
        """释放资源

        :param session_id: 会话ID
        :type session_id: str
        :param res_type: 资源类型
        :type res_type: str
        :param resource_id: 资源ID
        :type resource_id: str
        """
        raise NotImplementedError()
    
    def get_file(self, path):
        """获取一个文件资源

        :param path: 相对于resources目录的文件路径
        :type path: str
        :returns: 文件绝对路径
        :rtypes: str
        """
        raise NotImplementedError()
    
    def list_dir(self, path):
        """获取一个文件资源

        :param path: 相对于resources目录的文件夹路径
        :type path: str
        :returns: 文件绝对路径
        :rtypes: str
        """
        raise NotImplementedError()
    
    def walk(self, path):
        """遍历一个文件路径

        :param path: 相对于resources目录的文件夹路径
        :type path: str
        :returns: 返回一个迭代器，每次迭代对应一个(dir_path, dir_names, file_names)的元组
        :rtypes: iterator
        """
        raise NotImplementedError()
    
    def iter_managed_resource(self):
        """查询全部托管的资源（支持初始化&反初始化）

        :returns: iterator of (res_type, resource)
        :rtypes: iterator(res_type, resource)
        """
        raise NotImplementedError()
    
    
class TestResourceManager(object):
    """测试资源管理"""
    def __init__(self, backend):
        self._backend = backend
        
    def create_session(self, testcase=None):
        """创建资源使用会话

        :param testcase: 使用的测试用例
        :type testcase: TestCase
        :returns: 会话
        :rtype: Session
        """
        session_id = self._backend.create_session(testcase)
        return Session(self._backend, session_id)

    def iter_managed_resource(self):
        """查询全部托管的资源（支持初始化&反初始化）

        :returns: iterator of (res_type, resource)
        :rtypes: iterator(str, dict)
        """
        return self._backend.iter_managed_resource()

    
class LocalResourceLock(object):
    """本地资源锁
    """

    _lock_cache = {}

    def __init__(self, res_type, resource_id):
        self._res_type = res_type
        self._resource_id = resource_id
        if sys.platform == "win32":
            lock_dir = os.path.join(os.environ["AppData"], "QTAF", "lock", res_type)
            self._os_try_acquire = self._win_try_acquire
            self._os_release = self._win_release
        else: #linux/mac
            lock_dir = os.path.join(os.environ["HOME"], ".qtaf", "lock", res_type)
            self._os_try_acquire = self._unix_try_acquire
            self._os_release = self._unix_release
        if not os.path.isdir(lock_dir):
            os.makedirs(lock_dir)
        self._file_path = os.path.join(lock_dir, str(resource_id))
        self._fd = None

    def try_acquire(self):
        """尝试加锁

        :returns: 是否成功
        """
        if self._file_path in self._lock_cache:
            return False
        if self._os_try_acquire():
            self._lock_cache[self._file_path] = self
            return True
        return False

    def release(self):
        """释放
        """
        if self._fd is None:
            raise RuntimeError("lock is not acquired")
        del self._lock_cache[self._file_path]
        return self._os_release()

    def _win_try_acquire(self):
        import msvcrt
        try:
            fd = os.open(self._file_path, os.O_RDWR | os.O_CREAT | os.O_TRUNC)
        except OSError:
            pass
        else:
            try:
                msvcrt.locking(fd, msvcrt.LK_NBLCK, 1)
            except (IOError, OSError):
                os.close(fd)
            else:
                self._fd = fd
                return True
        return False

    def _win_release(self):
        import msvcrt
        msvcrt.locking(self._fd, msvcrt.LK_UNLCK, 1)
        os.close(self._fd)
        try:
            os.remove(self._file_path)
        except OSError:
            pass
        self._fd = None

    def _unix_try_acquire(self):
        import fcntl
        fd = os.open(self._file_path, os.O_RDWR | os.O_CREAT | os.O_TRUNC)
        try:
            fcntl.flock(fd, fcntl.LOCK_EX | fcntl.LOCK_NB)
        except (IOError, OSError):
            os.close(fd)
        else:
            self._fd = fd
            return True
        return False

    def _unix_release(self):
        import fcntl
        fcntl.flock(self._fd, fcntl.LOCK_UN)
        os.close(self._fd)
        self._fd = None

class LocalResourceHandler(object):
    """本地资源处理句柄
    """
    def __init__(self, resource_lock_type=LocalResourceLock ):
        """构造函数

        :param resource_lock_type: 资源锁类型
        :type resource_lock_type: type
        """
        self._lock_type = LocalResourceLock
        self._res_locks = {}
        
    def acquire_resource(self, session_id, res_type, res_group, condition):
        """申请资源
    
        :param session_id: 会话ID
        :type session_id: str
        :param res_type: 资源类型
        :type res_type: str
        :param res_group: 资源分组
        :type res_group: str
        :param condition: 资源属性匹配
        :type condition: dict
        :returns: 资源
        :rtype: dict
        """
        for it in self.iter_resource(res_group, condition):
            if not isinstance(it, dict):
                raise ValueError("Resource record should be a dictionary")
            if "id" not in it:
                raise ValueError("Resource record should be a dictionary with key 'id'")
            resource_id = it["id"]
            lock = self._lock_type(res_type, resource_id)
            if lock.try_acquire():
                self._res_locks.setdefault(session_id, [])
                self._res_locks[session_id].append((res_type, resource_id, lock))
                return it
        else:
            raise ResourceNotAvailable()

    def release_resource(self, session_id, res_type, resource_id):
        """释放资源

        :param session_id: 会话ID
        :type session_id: str
        :param res_type: 资源类型
        :type res_type: str
        :param resource_id: 资源ID
        :type resource_id: str
        """
        if session_id not in self._res_locks:
            raise ValueError("invalid session")
        for idx, (_res_type, _resource_id, lock) in enumerate(self._res_locks[session_id]):
            if _res_type == res_type and _resource_id == resource_id:
                lock.release()
            del self._res_locks[session_id][idx]
            break
        else:
            raise ValueError("resource is not acquired yet")

    def session_created(self, session_id, timeout, testcase=None):
        """通知会话创建

        :param session_id: 会话ID
        :type session_id: str
        :param timeout: 会话超时时间（秒）
        :type timeout: int
        :param testcase: 当前用例
        :type testcase: testbase.testcase.TestCase
        """
        pass

    def session_destroyed(self, session_id):
        """通知会话销毁

        :param session_id: 会话ID
        :type session_id: str
        """
        if session_id in self._res_locks:
            for _, _, lock in self._res_locks[session_id]:
                lock.release()
            del self._res_locks[session_id]

    def iter_resource(self, res_type, res_group=None, condition=None):
        """遍历全部资源（可以按照优先级顺序返回来影响申请资源的优先级）

        :param res_type: 资源类型
        :type res_type: str
        :param res_group: 资源分组
        :type res_group: str
        :param condition: 资源属性匹配
        :type condition: dict
        :returns: iterator of resource, dict type with key 'id'
        :rtypes: iterator(dict)
        """
        raise NotImplementedError()

    def iter_managed_resource(self, res_type):
        """查询全部托管的资源（支持初始化&反初始化）

        :returns: iterator of resource
        :rtypes: iterator(dict)
        """
        return self.iter_resource(res_type)

class LocalCSVResourceHandler(LocalResourceHandler):
    """基于本地CSV文件管理资源
    """
    def __init__(self, csv_path, resource_lock_type=LocalResourceLock ):
        """构造函数

        :param csv_path: CSV文件
        :type csv_path: str
        :param resource_lock_type: 资源锁类型
        :type resource_lock_type: type
        """
        self._csv_path = csv_path
        super(LocalCSVResourceHandler, self).__init__(resource_lock_type)
    
    def iter_resource(self, res_type, res_group=None, condition=None):
        """遍历全部资源（可以按照优先级顺序返回来影响申请资源的优先级）

        :param res_type: 资源类型
        :type res_type: str
        :param res_group: 资源分组
        :type res_group: str
        :param condition: 资源属性匹配
        :type condition: dict
        :returns: iterator of resource
        :rtypes: iterator(dict)
        """
        with codecs_open(self._csv_path, "rb", encoding="utf-8") as fd:
            for rowid, row in enumerate(csv.DictReader(fd)):
                if "id" not in row:
                    row["id"] = rowid
                    yield row

class LocalResourceManagerBackend(IResourceManagerBackend):
    """基本本地文件的方式的资源管理
    """
    _res_type_map = {}
        
    @classmethod
    def register_resource_type(cls, res_type, handler):
        """注册一个资源类型
        """
        cls._res_type_map[res_type] = handler

    def __init__(self):
        self._resources_dirs = iter_resource_paths()
        self._session_path = os.path.join(settings.PROJECT_ROOT, 'sessions')
    
    def _adjust_path(self,path):
        """根据操作系统转换文件分隔符
        """
        if os.sep == '/':
            return path.replace('\\',os.sep)
        else:
            return path.replace('/',os.sep)
    
    def _download_file(self, url, target_path):
        from six.moves.urllib import request, error
        try:
            rsp = request.urlopen(url, timeout=300)
            rspbuf = rsp.read()
        except error.HTTPError as e:
            raise DownloadFileError(url, e.code, e.msg, e.headers, e.read())
        with codecs_open(target_path, "wb") as fd:
            fd.write(rspbuf)
    
    def _resolve_link_file(self, remote_path, prefer_local_path):
        """获取链接的真正的文件
        """
        if os.path.isfile(remote_path):
            return remote_path
        elif remote_path.startswith("http://") or remote_path.startswith("https://"):
            self._download_file(remote_path, prefer_local_path)
            return prefer_local_path
        else:
            raise ValueError("Invalid link file path: %s" % remote_path)

    def get_file(self,relative_path):
        """查找某个文件

        :type relative_path:string
        :param relative_path: ，资源文件相对描述符，相对于setting下的资源目录的路径,支持多级目录
        :return:返回资源文件的绝对路径
        """
        result = []
        relative_path = self._adjust_path(relative_path)
        if relative_path.startswith(os.sep):
            relative_path = relative_path[1:]
        for it in self._resources_dirs:
            file_path = self._adjust_path(os.path.join(it,relative_path))
            file_path = smart_text(file_path)
            file_link = smart_text(file_path+'.link')
            if os.path.isfile(file_path):
                result.append(file_path)
            elif os.path.isfile(file_link):
                with codecs_open(file_link, encoding="utf-8") as f:
                    remote_path = f.read()
                    file_path = self._resolve_link_file(remote_path,file_path)
                    result.append(file_path)
        if len(result) > 1:
            raise Exception("存在多个%s文件" %relative_path)
        elif len(result) < 1:
            raise Exception("%s文件不存在" %relative_path)
        return result[0]
    
    def list_dir(self,relative_path):
        """列出某个目录下的文件

        :type relative_path:string
        :param relative_path: ，资源文件目录相对路径，相对于setting下的资源目录的路径,支持多级目录
        :return:返回一个包含资源目录下所有文件或者文件下的绝对路径列表
        """
        result = []
        relative_path = self._adjust_path(relative_path)
        if relative_path.startswith(os.sep):
            relative_path = relative_path[1:]
        for it in self._resources_dirs:
            dir_path = self._adjust_path(os.path.join(it,relative_path))
            dir_path = smart_text(dir_path)
            if os.path.isdir(dir_path):
                result.append(dir_path)
        if len(result) > 1:
            logger.error("找到多个目录:")
            for item in result:
                logger.error("%s" %item)
            raise Exception("存在多个%s目录" %relative_path)
        if len(result) < 1:
            raise Exception("%s目录不存在" %relative_path)
        paths=[]
        for path in os.listdir(result[0]):
            paths.append(os.path.join(result[0],path))
        return paths
    
    def walk(self, path):
        """获取目录下文件列表

        :param path: 相对于resources目录的路径，用于遍历文件夹
        :type  path: str
        :return iterators: iterator of (dir_path, dirnames, filenames) tuples
        :type   iterators: iterator
        """
        sub_dirs = [path]
        while sub_dirs:
            sub_dir = sub_dirs.pop()
            items = self.list_dir(sub_dir)
            dir_names = []
            file_names = []
            for item in items:
                base_item = os.path.basename(item)
                if base_item.endswith(".link"):
                    base_item = base_item[:-5]
                if os.path.isdir(item):
                    dir_names.append(base_item)
                    sub_dirs.append(base_item)
                else:
                    file_names.append(base_item)
            yield sub_dir, dir_names, file_names
        
    def _clean_cache(self):
        """清理缓存文件
        """
        for it in self._resources_dirs:
            self._rm_cachefile_recursively(it)
    
    def _rm_cachefile_recursively(self,path):
        """递归删除目录下的缓存文件
        """
        for root, _, files in os.walk(path):
            for it in files:
                if it.endswith('.link'):
                    try:
                        os.remove(os.path.join(root,it)[0:-5])
                    except OSError:
                        pass
                    
    def create_session(self, testcase=None):
        """创建会话
        """
        session_id = str(uuid.uuid4())
        timeout = 300 if testcase is None else (testcase.timeout * 60 + 300 + 5)  # 300为用例超时时的Cleanup超时时间
        for handler in self._res_type_map.values():
            handler.session_created(session_id, timeout, testcase)
        return session_id

    def destroy_session(self, sessionid):
        """销毁会话
        """
        self._clean_cache()
        for handler in self._res_type_map.values():
            handler.session_destroyed(sessionid)
    
    def acquire_resource(self, session_id, res_type, res_group, condition ):
        """申请资源
        """
        handler = self._res_type_map.get(res_type)
        if handler is None:
            raise ValueError("resource type '%s' it not registered")
        return handler.acquire_resource(session_id, res_type, res_group, condition)
        
    def release_resource(self, session_id, res_type, resource_id):
        """释放资源
        """
        handler = self._res_type_map.get(res_type)
        if handler is None:
            raise ValueError("resource type '%s' it not registered")
        return handler.release_resource(session_id, res_type, resource_id)

    def iter_managed_resource(self):
        """查询全部托管的资源（支持初始化&反初始化）

        :returns: iterator(res_type, resource)
        """
        for res_type, handler in self._res_type_map.items():
            for it in handler.iter_managed_resource(res_type):
                yield (res_type, it)


def _current_resmgr_session():
    tc = context.current_testcase()
    if tc is None:
        logger.warn("注意！非用例模式，将返回默认测试文件资源管理器，此处调用可能在测试环境下异常")
        return TestResourceManager(LocalResourceManagerBackend()).create_session()
    return tc.test_resources

        
def get_file(path):
    """查找某个文件
    :param path: 相对于resources目录的路径，用于查找文件
    :type path: str
    :return:返回资源文件的绝对路径
    """
    return _current_resmgr_session().get_file(path)

def list_dir(path):
    """列出某个目录下的文件
    :param path: 相对于resources目录的路径，用于查找文件夹
    :type path: str
    :returns :返回一个包含资源目录下所有文件或者文件下的绝对路径的list
    """
    return _current_resmgr_session().list_dir(path)

def walk(path):
    """遍历某个路径
    
    :param path: 相对于resources目录的路径，用于遍历文件夹
    :type  path: str
    :return iterators: iterator of (dir_path, dirnames, filenames) tuples
    :type   iterators: iterator
    """
    return _current_resmgr_session().walk(path)

def acquire_resource(res_type, res_group=None, condition=None):
    """申请资源

    :param res_type: 资源类型
    :type res_type: str
    :param res_group: 资源分组
    :type res_type: str
    :param condition: 资源属性匹配
    :type condition: dict
    :returns: 资源
    :rtypes: dict
    """
    return _current_resmgr_session().acquire_resource(res_type, res_group, condition)

def release_resource(res_type, resource_id):
    """释放资源

    :param res_type: 资源类型
    :type res_type: str
    :param resource_id: 资源ID
    :type resource_id: str
    """
    return _current_resmgr_session().release_resource(res_type, resource_id)

def iter_resource_paths():
    """返回测试项目的全部resources目录

    :return: 
    """
    resource_paths = []
    for dirpath, dirnames, _ in os.walk(settings.PROJECT_ROOT):
        for dirname in dirnames:
            if dirname == 'resources':
                resource_paths.append(os.path.join(dirpath, dirname))
    return resource_paths


def __init_resmgr_backend_types():
    global resmgr_backend_types
    if resmgr_backend_types:
        return
    resmgr_backend_types["local"] = LocalResourceManagerBackend
    for ep in pkg_resources.iter_entry_points(RESMGR_BACKEND_ENTRY_POINT):
        if ep.name not in resmgr_backend_types:
            resmgr_backend_types[ep.name] = ep.load()

__init_resmgr_backend_types()
del __init_resmgr_backend_types



# -*- coding: utf-8 -*-
#
# Tencent is pleased to support the open source community by making QTA available.
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the BSD 3-Clause License (the "License"); you may not use this 
# file except in compliance with the License. You may obtain a copy of the License at
# 
# https://opensource.org/licenses/BSD-3-Clause
# 
# Unless required by applicable law or agreed to in writing, software distributed 
# under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#
'''
测试用例基类模块
'''

import os
import sys
import re
import threading
import traceback
import collections
import random
import types
import six

from testbase.assertion import AssertionRewriter
from testbase.util import Singleton, ThreadGroupLocal, ThreadGroupScope, smart_text, get_last_frame_stack
from testbase.testresult import EnumLogLevel, TestResultCollection
from testbase.conf import settings
from testbase.retry import Retry


# 后续需专门花时间去除TestCaseStatus和TestCasePriority这两个类
class TestCaseStatus(object):
    '''测试用例状态
    
    :attention: 此类将会被移除，请使用TestCase.EnumStatus
    '''
    Design, Implement, Review, Ready, Suspend = ('Design', 'Implement', 'Review', 'Ready', 'Suspend')

class TestCasePriority(object):
    '''测试用例优先级
    
    :attention: 此类将会被移除，请使用TestCase.EnumPriority
    '''
    BVT, High, Normal, Low = ('BVT', 'High', 'Normal', 'Low')


class Environ(six.with_metaclass(Singleton, dict)):
    """测试环境类

用法说明：
它是一个继承了字典类型的单实例类。使用时，需先实例化该类，如：
from testbase.testcase import Environ
env = Environ()
env保存了用例运行时的一些测试环境变量。

测试环境变量分为3个部分：

1、在测试执行时，env则存储了由测试计划定义的用例环境变量，使用方法如下::

    from testbase.testcase import Environ
    env = Environ()
    print(env_
    # 输出
    env = {
        'ASSERTTEST':'True',  #注意：这里key是全字母大写
    }

2、测试用例基类testbase.testcase.TestCase的构造函数中也实例化了Environ类（变量名为environ)，
并且，保存了当时执行用例类的类名和类说明，使用方法如下::

    #todo: 执行用例中使用evniron，用于打印用例名称和用例说明
    from testbase.testcase import TestCase
    class YourTest(TestCase):
        def runTest(self):
            print(self.environ['TestName'])
            print(self.environ['TestDoc'])

3、Environ还可以用于设置自定义的环境变量，使用方法如下::

    from testbase.testcase import Environ
    env = Environ()
    env['YourEnvKey'] = "EnvValue" 
    print(env['YourEnvKey'])

    """
    def __init__(self):
        """构造函数。判断系统环境变量中是否存在由测试计划中传入的环境变量，有则加载。
        
        """
        super(Environ, self).__init__()
        for key in os.environ.keys():
            if key.startswith("QTAF_") and key[5:]:
                self[key[5:]] = os.environ[key]


class TestCaseType(type):
    '''测试用例元类型
    '''
    forbid_overload_methods = ["__init__"]

    def __new__(cls, name, bases, attrs):
        super_new = super(TestCaseType, cls).__new__
        parents = [b for b in bases if isinstance(b, TestCaseType)]
        # ensure only apply for subclass of TestCase
        if not parents:
            return super_new(cls, name, bases, attrs)

        for it in cls.forbid_overload_methods:
            if it in attrs.keys():
                raise RuntimeError("不允许%s重载函数: %s" % (cls.__name__, it))

        base_tags_set = set()
        if "tags" in attrs:
            tags = attrs.pop("tags")
            if isinstance(tags, six.string_types):
                tags = [tags]
            tags_set = set(tags)
            for b in bases:
                if issubclass(b, TestCase) and hasattr(b, "tags"):
                    base_tags_set |= b.tags
            if "__module__" in attrs:
                mod = sys.modules[attrs["__module__"]]
                if hasattr(mod, "__qtaf_tags__"):
                    mod_tags = mod.__qtaf_tags__
                    if isinstance(mod_tags, six.string_types):
                        mod_tags = [mod_tags]
                    base_tags_set |= set(mod_tags)
            tags_set |= base_tags_set
        else:
            tags_set = set()
        attrs["tags"] = tags_set
        attrs["base_tags"] = base_tags_set
        return super_new(cls, name, bases, attrs)

@six.add_metaclass(TestCaseType)
class TestCase(object):
    '''测试用例基类
    
            所有测试用例都最终从此基类继承。测试用例的测试脚本主要实现在"runTest()"中，
           而当用例需要初始化和清理测试环境时则分别重写"preTest()"和"postTest()"函数。
    '''
    test_extra_info_def = [] #自定义字段
    
    class EnumStatus(object):
        '''测试用例状态枚举类
        
        :attention: 如果因为特殊原因需要暂时屏蔽某个用例的任务执行（比如有功能缺陷从而导致执行失败），
                                                     则可以先置为该字段为Suspend,等到可用的时候再将该字段置为Ready
        '''
        Design, Implement, Review, Ready, Suspend = (TestCaseStatus.Design, 
                                                     TestCaseStatus.Implement,
                                                     TestCaseStatus.Review,
                                                     TestCaseStatus.Ready,
                                                     TestCaseStatus.Suspend)
        
    class EnumPriority(object):
        '''测试用例优先级枚举类
        '''
        BVT, High, Normal, Low = (TestCasePriority.BVT,
                                  TestCasePriority.High,
                                  TestCasePriority.Normal,
                                  TestCasePriority.Low)
        
    owner = None
    priority = None
    status = None
    timeout = None

    ATTRIB_OVERWRITE_WHITELIST = ["priority", "status", "owner", "timeout", "tags", "__doc__"]

    def __init__(self, testdata=None, testdataname=None, attrs=None ):
        '''构造函数 
        
        :param testdata: 测试数据
        :type testdata: object
        :param testdataname: 测试数据标识
        :type testdataname: str
        :param attrs: 重载的测试用例属性
        :type attrs: dict
        '''
        self.__casedata = testdata
        self.__testdataname = testdataname
        self.__environ = Environ()
        self.__testresult = None
        self.__resmgr_session = None
        self.__resmgr = None
        self.__test_doc = None

        if attrs:
            for k, v in attrs.items():
                if k in self.ATTRIB_OVERWRITE_WHITELIST:
                    if k == "__doc__":
                        self.__test_doc = v
                    elif k == "tags":
                        if isinstance(v, six.string_types):
                            v = [v]
                        self.tags = self.base_tags | set(v)
                    else:
                        setattr(self, k, v)

    @property
    def casedata(self):
        '''测试数据
        
        :rtype: list
        '''
        return self.__casedata
    
    @property
    def casedataname(self):
        '''测试数据标识
        
        :rtype: str
        '''
        return self.__testdataname
    
    @property
    def environ(self):
        '''环境变量
        
        :rtype: Environ
        '''
        return self.__environ
    
    @property
    def test_result(self):
        '''对应的测试结果
        
        :rtype: TestResult
        '''
        return self.__testresult
    
    @property
    def test_dir(self):
        '''测试用例执行的临时目录
        
        :rtype: str
        '''
        return os.path.abspath(os.getcwd())
        
    @property
    def test_class_name(self):  
        '''返回测试用例名字（不同测试用例的名字不同）
        
        :rtype: str
        '''
        cls = type(self)
        if cls.__module__ == '__main__':
            type_name =  smart_text(cls.__name__)
        else:
            type_name = smart_text(cls.__module__ + '.' + cls.__name__)
        return type_name
                        
    @property
    def test_name(self):
        '''返回测试用例实例的名字
        
        :rtype: str
        '''
        if self.casedata is not None:
            casedataname = smart_text(self.casedataname)
            return '%s/%s' % (self.test_class_name, casedataname)
        else:
            return self.test_class_name
        
    @property
    def test_doc(self):
        '''测试用例说明
        
        :rtype: str
        '''
        if self.__test_doc:
            return self.__test_doc
        desc = self.__class__.__doc__
#        if desc:
#            desc = cgi.escape(desc)
        if isinstance(desc, six.text_type):
            desc = re.sub('^\s*', '', desc)
            desc = re.sub('\s*$', '', desc)
        return desc
    
    @property
    def test_extra_info(self):
        '''测试用例额外信息
        '''
        info = {}
        for name, _ in self.test_extra_info_def:
            info[name] = getattr(self, name, None)
        return info
    
    @property
    def test_resmgr(self):
        '''资源管理器
        '''
        return self.__resmgr
    
    @test_resmgr.setter
    def test_resmgr(self, resmgr):
        self.__resmgr = resmgr

    @property
    def test_resources(self):
        '''资源管理使用接口
        '''
        if not self.__resmgr_session:
            self.__resmgr_session = self.__resmgr.create_session(self)
        return self.__resmgr_session            
    
    def init_test(self, testresult ):
        '''初始化测试用例。慎用此函数，尽量将初始化放到preTest里。
        
        :param testresult: 测试结果
        :type testresult: TestResult
        '''
        self.__testresult = testresult
        
    def pre_test(self):
        '''测试环境初始化
        '''
        pass
    
    def run_test(self):
        '''运行测试用例
        '''
        raise NotImplementedError("请在%s类中实现runTest方法" % type(self))

    def post_test(self):
        '''测试环境清理
        '''
        pass
    
    def clean_test(self):
        '''测试用例反初始化。慎用此函数，尽量将清理放到postTest里。
        '''
        self.test_resources.destroy()
    
    def start_step(self, stepinfo):
        '''开始执行一个测试步骤
        
        :param stepinfo: 步骤描述
        :type stepinfo: str
        '''
        if not isinstance(stepinfo, six.text_type):
            stepinfo=str(stepinfo)
        self.__testresult.begin_step(stepinfo)
        
    def log_info(self, info ):
        '''Log一条信息
        
        :type info: string
        :param info: 要Log的信息  
        '''
        if not isinstance(info, six.text_type):
            info=str(info)
        self.__testresult.info(info)
        
    def fail(self, message):
        '''测试用例失败
        
        :type message: string
        :param message: 要Log的信息  
        '''
        if not isinstance(message, six.text_type):
            message=str(message)
        self.__testresult.error(message)
        
    def __record_assert_failed( self, message, actual, expect ):
        '''记录Assert失败信息
        
        :param message: 提示信息
        :type message: string
        :param actual: 实际值
        :type actual: string
        :param expect: 期望值
        :type expect: string
        '''
        #得到上一个函数调用帧所在的文件路径，行号，函数名
        stack = get_last_frame_stack(3)
        msg = "检查点不通过\n%s%s\n期望值：%s%s\n实际值：%s%s" % (smart_text(stack), smart_text(message),
                                                    expect.__class__, expect,
                                                    actual.__class__, actual)
        self.__testresult.log_record(EnumLogLevel.ASSERT, msg)
    
    def _log_assert_failed(self, message, back_count=2):
        """记录断言失败的信息
        """
        stack = get_last_frame_stack(back_count)
        msg = "检查点不通过\n%s%s\n" % (smart_text(stack), smart_text(message))
        self.__testresult.log_record(EnumLogLevel.ASSERT, msg)
        
    def assert_(self, message, value):
        """测试断言，如果value的值不为真，则用例失败，输出对应信息
        
        :param message:断言失败时的提示消息
        :type  message: str
        :param value:用于判断的值
        :type  value: bool或
        """
        if not value:
            self._log_assert_failed(message, 3)
    
    def assert_equal(self, message, actual, expect=True):
        '''检查实际值和期望值是否相等，不能则测试用例失败
        
       :param message: 检查信息
       :param actual: 实际值
       :param expect: 期望值(默认：True)
       :return: True or False
        '''
        if isinstance(actual, six.string_types):
            actual = smart_text(actual)
        if isinstance(expect, six.string_types):
            expect = smart_text(expect)
        if expect != actual:
            self.__record_assert_failed(message, actual, expect)
            return False
        else:
            return True
    
    def assert_match(self, message, actual, expect):
        '''检查actual和expect是否模式匹配，不匹配则记录一个检查失败
        
        :type message: string
        :param message: 失败时记录的消息
        :type actual: string
        :param actual: 需要匹配的字符串
        :type expect: string
        :param expect: 要匹配的正则表达式 
        :return: 匹配成果
        '''
        if isinstance(actual,six.string_types):
            actual = smart_text(actual)
        if isinstance(expect,):
            expect = smart_text(expect)        
        if re.search(expect, actual):
            return True
        else:
            self.__record_assert_failed(message, actual, expect)
            return False
    
    def wait_for_equal(self, message, obj, prop_name, expected, timeout=10, interval=0.5):
        '''每隔interval检查obj.prop_name是否和expected相等，如果在timeout时间内都不相等，则测试用例失败

        :param message: 失败时的输出信息
        :param obj: 需要检查的对象
        :type prop_name: string 
        :param prop_name: 需要检查的对象的属性名，支持多层属性
        :param expected: 期望的obj.prop_name值
        :param timeout: 超时秒数
        :param interval: 重试间隔秒数
        :return: True or False
        '''
        for _ in Retry(timeout=timeout, interval=interval, raise_error=False):
            actual = getattr(obj, prop_name)
            if actual == expected:
                return True
        else:
            self.__record_assert_failed(message, getattr(obj, prop_name), expected)
            return False
        
        
    def wait_for_match(self, message, obj, prop_name, expected, timeout=10, interval=0.5):
        '''每隔interval检查obj.prop_name是否和正则表达式expected是否匹配，如果在timeout时间内都不相等，则测试用例失败

        :param message: 失败时的输出信息
        :param obj: 需要检查的对象
        :type prop_name: string 
        :param prop_name: 需要检查的对象的属性名, obj.prop_name返回字符串
        :param expected: 需要匹配的正则表达式
        :param timeout: 超时秒数
        :param interval: 重试间隔秒数
        :return: True or False
        '''
        for _ in Retry(timeout=timeout, interval=interval, raise_error=False):
            actual = getattr(obj, prop_name)
            if re.match(expected, actual, re.I):
                return True
        else:
            self.__record_assert_failed(message, getattr(obj, prop_name), expected)
            return False
            
    def get_extra_fail_record(self):
        '''当错误发生时，获取需要额外添加的日志记录和附件信息
        
        :rtype: dict,dict - 日志记录，附件信息
        '''
        return {}, {}
    
    def debug_run(self):
        '''本地调试测试用例
        '''
        from testbase.runner import TestRunner
        from testbase.loader import TestDataLoader
        from testbase.report import EmptyTestReport, StreamTestReport
        from testbase.testresult import StreamResult
        from testbase import datadrive
        
        if self.casedata is None:
            testcls = type(self)
            if datadrive.is_datadrive(testcls):  #数据驱动用例
                runner = TestRunner(StreamTestReport(output_testresult=True, output_summary=True))
                return runner.run(datadrive.load_datadrive_tests(testcls))
            elif settings.DATA_DRIVE: #项目级的数据驱动
                testdataset = TestDataLoader().load()
                runner = TestRunner(StreamTestReport(output_testresult=True, output_summary=True))
                return runner.run([testcls(testdataset[it], str(it)) for it in testdataset ])
        runner = TestRunner(EmptyTestReport(lambda tc: StreamResult()))
        return runner.run([self])
     
    def debug_run_one(self, name=None ):
        '''本地调试测试用例，给数据驱动的用例使用，只执行一个用例
        
        :param name: 测试数据名称，如果不指定，执行第一个数据的用例
        '''
        from testbase.runner import TestRunner
        from testbase.loader import TestDataLoader
        from testbase.report import EmptyTestReport
        from testbase.testresult import StreamResult
        from testbase import datadrive

        runner = TestRunner(EmptyTestReport(lambda tc: StreamResult()))
        if self.casedata is None:
            testcls = type(self)
            if datadrive.is_datadrive(testcls):  #数据驱动用例
                tests = datadrive.load_datadrive_tests(testcls, name)
                if len(tests) > 1:
                    tests = [ random.choice(tests) ]
                return runner.run(tests)
            elif settings.DATA_DRIVE: #项目级的数据驱动
                testdataset = TestDataLoader().load()
                for it in testdataset:
                    if (name is None) or (str(name) == str(it)):
                        return runner.run([testcls(testdataset[it], str(it))])
                else:
                    raise RuntimeError("找不到指定名字的测试数据")
            else:
                raise RuntimeError("非数据驱动用例请使用debug_run接口进行调试执行")
        else:
            runner.run([self])
            return self
        
        
        
    #----------------------------------------------------   
    #    以下为兼容老的代码风格的接口，新代码请勿再使用
    #----------------------------------------------------
    
    def run(self):
        '''本地调试测试用例
        '''
        self.debug_run()
        
        sys.stderr.write('============================================================\n')
        sys.stderr.write('注意：TestCase.run接口准备废弃，建议改用为.debug_run接口，例如：\n')
        sys.stderr.write('if __name__ == \'__main__\'\n')
        sys.stderr.write('    %s().debug_run()\n' % type(self).__name__)
        sys.stderr.write('============================================================\n')
        
        
    initTest = init_test
    preTest = pre_test
    runTest = run_test
    postTest = post_test
    cleanTest = clean_test
    startStep = start_step
    
    assertEqual = assert_equal
    assertMatch = assert_match
    waitForEqual = wait_for_equal
    waitForMatch = wait_for_match
    
    TestClassName = test_class_name
    TestName = test_name
    TestDoc = test_doc
    
    logInfo = log_info
        
class ITestCaseRunner(object):
    '''测试用例执行器接口定义
    '''
    def run(self, testcase, testresult_factory):
        '''执行一个测试用例
        
        :param testcase: 执行的测试用例
        :type testcase: TestCase
        :param testresult_factory: 测试结果工厂
        :type testresult_factory: ITestResultFactory
        :return TestResult/TestResultCollection - 测试结果
        '''
        raise NotImplementedError()
        
class TestCaseRunner(ITestCaseRunner):
    '''负责执行一个测试用例
    
    如果一个测试用例没有指定case_runner类变量，则默认都使用TestCaseRunner来执行这个用例。
    测试用例可以自定义和TestCaseRunner接口兼容的runner类，并设置case_runner类变量来实现
    自定义一个测试用例的执行逻辑，以下是TestCaseRunner的接口定义
    '''
    
    
    CLEANUP_TIMEOUT = 300
    
    def __init__(self):
        '''构造函数
        '''
        self._lock = threading.Lock()
        self._stop_run = False
        self._error = None
        
    def _rewrite_assert(self, cls):
        if not settings.get("QTAF_REWRITE_ASSERT", True) or cls == TestCase:
            return
        rewriter = AssertionRewriter()
        for key in dir(cls):
            item = getattr(cls, key)
            if isinstance(item, (types.MethodType, types.FunctionType)):
                rewriter.rewrite(item)
        
    def _walk_bases(self,test_case):
        '''遍历所有基类，进行相应的处理
        '''
        cur_cls=type(test_case)
        self._single_methods_mapping(cur_cls)
        self._rewrite_assert(cur_cls)
        bases=set(cur_cls.__bases__)
        while bases:
            new_bases=set()
            for base_cls in bases:
                self._single_methods_mapping(base_cls)
                self._rewrite_assert(cur_cls)
                new_bases=new_bases.union(set(base_cls.__bases__))
            bases=new_bases
        
    def _single_methods_mapping(self,cls_type):
        '''针对单个类进行方法映射
        '''
        if object==cls_type:
            return
        case_mothods=[
            ("initTest","init_test"),
            ("preTest","pre_test"),
            ("runTest","run_test"),
            ("postTest","post_test"),
            ("cleanTest","clean_test")
        ]
        cls_dict=cls_type.__dict__
        for pairs in case_mothods:
            if pairs[0] in cls_dict and not pairs[1] in cls_dict:
                setattr(cls_type,pairs[1],cls_dict[pairs[0]])
                self._testresult.warning("严重警告：类型%s应当定义方法%s，方法%s已被废弃，请勿使用" % (cls_type.__name__,pairs[1],pairs[0]))             
            elif pairs[0] not in cls_dict and pairs[1] in cls_dict:
                setattr(cls_type,pairs[0],cls_dict[pairs[1]])
            elif pairs[0] in cls_dict and pairs[1] in cls_dict:
                if cls_dict[pairs[0]] != cls_dict[pairs[1]]:
                    raise RuntimeError('类型%s不允许同时独立定义"%s"和"%s"' % (cls_type,pairs[0],pairs[1]))
        
    def _check_testcase(self, testcase ):
        '''检查测试用例
        
        :param testcase: 测试用例
        :type testcase: TestCase
        :returns boolean, string - 检测是否通过，错误消息
        '''
        attrnames = ['owner', 'priority', 'status', 'timeout']
        tcattrs = {}
        for attr in attrnames:
            attrvalue = getattr(self._testcase, attr)
            if isinstance(attrvalue, six.text_type):
                if attrvalue.strip(" ") == "":
                    attrvalue = None
            tcattrs[attr] = attrvalue
        tcattrs['testname'] = self._testcase.test_name
        if None in tcattrs.values():
            raise RuntimeError("测试用例需要设置以下类属性：%s" % attrnames)
        if not self._testcase.test_doc:
            raise RuntimeError("测试用例的描述不能为空，请加上描述！")
        
        self._walk_bases(testcase)
        
    def _thread_run(self):
        '''测试用例线程过程
        '''
        #                     函数时发生了死锁，故注释掉。观察一段时间，看修改是否会影响测试。
        try:
            try:
                self._check_testcase(self._testcase)
            except RuntimeError as e:
                self._testresult.error(e.args[0])
                return
            
            while True:
                with self._lock:
                    if len(self._subtasks) == 0 or self._stop_run:
                        break             
                    it = self._subtasks.popleft()
                    
                if isinstance(it, str):
                    try:
                        if it in ['init_test', 'initTest']:
                            getattr(self._testcase, it)(self._testresult)
                        else:
                            getattr(self._testcase, it)()
                    except:
                        self._testresult.exception('%s执行失败' % it)
                else:
                    it()

        except:
            self._error = traceback.format_exc()
            
    def _thread_cleanup(self):
        '''清理线程
        '''
        try:
            while True:
                with self._lock:
                    if len(self._subtasks) == 0:
                        break             
                    it = self._subtasks.popleft()
                    if it in ['initTest', 'preTest', 'runTest', 'init_test', 'pre_test', 'run_test']:
                        continue
                    
                if isinstance(it, str):
                    try:
                        getattr(self._testcase, it)()
                    except:
                        self._testresult.exception('用例超时时%s执行失败' % it)
                else:
                    it()
        except:
            self._error = traceback.format_exc()
            
    def setup(self, testcase, testresult ):
        '''测试执行初始化
        
        :param testcase: 执行的测试用例
        :type testcase: TestCase
        :param testresult: 测试用例结果
        :type testresult: TestResult
        '''
        pass
    
    def teardown(self, testcase, testresult ):
        '''测试执行清理
        
        :param testcase: 执行的测试用例
        :type testcase: TestCase
        :param testresult: 测试用例结果
        :type testresult: TestResult
        '''
        pass
    
    def run(self, testcase, testresult_factory):
        '''执行一个测试用例
        
        :param testcase: 执行的测试用例
        :type testcase: TestCase
        :param testresult_factory: 测试结果工厂
        :type testresult_factory: ITestResultFactory
        :rtype: TestResult/TestResultCollection - 测试结果
        '''
        #                       临时方案是disable gc后kill掉程序，在enable gc。后续这里
        #                        需要重新考虑已独立进程来执行测试用例。
        
        self._stop_run=False
        self._testcase = testcase
        self._testresult = testresult_factory.create(testcase)
        self._subtasks = collections.deque(['init_test', 'pre_test', 'run_test', 'post_test', 'clean_test'])
                    
        with ThreadGroupScope('%s:%s'%(self._testcase.test_name, id(self))):

            ThreadGroupLocal().testcase = self._testcase
            ThreadGroupLocal().testresult = self._testresult
            
            self._testresult.begin_test(self._testcase)
            
            self.setup(self._testcase, self._testresult)
            
            if isinstance(self._testcase.timeout, int) or isinstance(self._testcase.timeout, float):
                timeout = self._testcase.timeout * 60
            else:
                timeout = 60
    
            test_thread = threading.Thread(target=self._thread_run)
            test_thread.daemon = True
            test_thread.start()
            test_thread.join(timeout)
            if test_thread.is_alive():
                self._stop_run = True
                try:
                    thread_traceback=self._get_current_traceback(test_thread)
                except:
                    self._testresult.log_record(EnumLogLevel.TESTTIMEOUT, '测试用例执行超时')
                else:
                    self._testresult.log_record(EnumLogLevel.TESTTIMEOUT, '测试用例执行超时，抓取测试线程当前堆栈', 
                                                dict(traceback=thread_traceback))
                
                #启动线程执行可能未执行的postTest和cleanTest
                cleanup_thread = threading.Thread(target=self._thread_cleanup)
                cleanup_thread.daemon = True
                cleanup_thread.start()
                cleanup_thread.join(self.CLEANUP_TIMEOUT)
                if cleanup_thread.is_alive():
                    try:
                        thread_traceback=self._get_current_traceback(cleanup_thread)
                    except:
                        self._testresult.log_record(EnumLogLevel.TESTTIMEOUT, '测试用例执行超时时清理超时')
                    else:
                        self._testresult.log_record(EnumLogLevel.TESTTIMEOUT, '测试用例执行超时时清理超时，抓取清理线程当前堆栈',
                                                    dict(traceback=thread_traceback))
                else:
                    if self._error:
                        raise RuntimeError("用例执行线程异常：\n%s" % smart_text(self._error))
            else:
                if self._error:
                    raise RuntimeError("用例执行线程异常：\n%s" % smart_text(self._error))
                    
            self.teardown(self._testcase, self._testresult)
            
            self._testresult.end_test()
            
        #gc.enable()
        #gc.collect()
        return self._testresult
        
    def _get_current_traceback(self, thread ):
        '''获取用例线程的当前的堆栈
        
        :param thread: 要获取堆栈的线程
        :type thread: Thread
        '''
        for thread_id, stack in sys._current_frames().items():
            if thread_id != thread.ident:
                continue
            tb = "Traceback ( thread-%d possibly hold at ):\n" % thread_id
            for filename, lineno, name, line in traceback.extract_stack(stack):
                tb += '  File: "%s", line %d, in %s\n' % (filename, lineno, name)
                if line:
                    tb += "    %s\n" % (line.strip())
            return tb
        else:
            raise RuntimeError("thread not found")
        
class RepeatTestCaseRunner(ITestCaseRunner):
    '''重复执行的用例执行器
    
    可以通过设置测试用例的类属性为此runner实例来实现指定测试用例
    执行多次。测试用例执行时可以访问成员变量iteration来判断当前是
    第几次执行。
    
    使用示例如下::
    
        class HelloRepeatTest(TestCase):
            '示例用例'
            case_runner = RepeatTestCaseRunner()
            owner = "foo"
            timeout = 1
            status = TestCase.EnumStatus.Ready
            priority = TestCase.EnumPriority.Normal
            
            def run_test(self):
                self.log_info("第%s次执行测试"%self.iteration)
    
    '''    
    
    def __init__(self, case_runner_class=None ):
        '''指定执行一个用例的CaseRunner类，不指定则为TestCaseRunner类
        '''
        if case_runner_class is None:
            case_runner_class = TestCaseRunner
        self._case_runner_class = case_runner_class
                
    def run(self, testcase, testresult_factory):
        '''执行一个测试用例
        
        :param testcase: 执行的测试用例
        :type testcase: TestCase
        :param testresult_factory: 测试结果工厂
        :type testresult_factory: ITestResultFactory
        :rtype TestResult/TestResultCollection - 测试结果
        '''
        passed = True
        if not hasattr(testcase, "repeat"):
            passed = False
            err_msg = "使用RepeatTestCaseRunner的测试用例需要设置类属性：repeat"
        elif testcase.repeat <= 0:
            passed = False
            err_msg = "使用RepeatTestCaseRunner的测试用例的类属性'repeat'必须大于0"
        if not passed:
            with ThreadGroupScope('%s:%s'%(testcase.test_name, id(self))):
                result = testresult_factory.create(testcase)
                ThreadGroupLocal().testcase = testcase
                ThreadGroupLocal().testresult = result
                result.begin_test(testcase)
                result.error(err_msg)
                result.end_test()
                return result
            
        passed = True
        results = []
        for i in range(testcase.repeat):
            testcase.iteration = i
            case_result = self._case_runner_class().run(testcase, testresult_factory)
            passed &= case_result.passed
            results.append(case_result)
            if not passed:
                break
        return TestResultCollection(results, passed)
        
class SeqTestCaseRunner(ITestCaseRunner):
    '''顺序执行的用例的执行器
    '''
    def run(self, testsuite, testresult_factory):
        '''执行一个顺序执行的测试用例套
        
        :param testsuite: 执行的测试用例套
        :type testsuite: SeqTestSuite
        :param testresult_factory: 测试结果工厂
        :type testresult_factory: ITestResultFactory
        :return TestResult/TestResultCollection - 测试结果
        '''
        passed = True
        results = []
        for it in testsuite:
            runner = getattr(it, 'case_runner', TestCaseRunner())
            case_result = runner.run(it, testresult_factory)
            passed &= case_result.passed
            results.append(case_result)
            if not passed:
                break
        return TestResultCollection(results, passed)
    
class TestSuite(object):
    '''测试用例套
    '''
    @property
    def suite_class_name(self):
        '''测试套类名称
        '''
        cls = type(self)
        if cls.__module__ == '__main__':
            type_name =  cls.__name__
        else:
            type_name = (cls.__module__ + '.' + cls.__name__)
        return type_name
        
    def dumps(self):
        '''序列化
        '''
        raise NotImplementedError()
    
    def loads(self, buf ):
        '''反序列化
        '''
        raise NotImplementedError()
        
class SeqTestSuite(TestSuite):
    '''顺序执行的测试用例套
    '''
    case_runner = SeqTestCaseRunner()
    
    def __init__(self, testcases ):
        '''构造函数
        
        :param testcases: 测试用例列表
        :type testcases: list
        :param name: 测试用例名
        :type name: string
        '''
        self._testcases = testcases
        self._resmgr = None
    
    def __iter__(self):
        for it in self._testcases:
            yield it
    
    def __repr__(self):
        return '<SeqTestSuite module:%s>' % self.test_class_name

    @property
    def test_class_name(self):  
        '''返回测试用例名字（不同测试用例的名字不同）
        
        :rtype: str
        '''
        cls = type(self._testcases[0])
        return cls.__module__
                        
    @property
    def test_name(self):
        '''返回测试用例实例的名字
        
        :rtype: str
        '''
        cls = type(self._testcases[0])
        return cls.__module__
        
    @property
    def test_doc(self):
        '''测试用例说明
        
        :rtype: str
        '''
        cls = type(self._testcases[0])
        desc = cls.__module__.__doc__
        if isinstance(desc, six.text_type):
            desc = re.sub('^\s*', '', desc)
            desc = re.sub('\s*$', '', desc)
        return desc
    
    @property
    def test_result(self):
        '''将最后一个执行的用例结果，作为Suite的结果
        '''
        result=None
        for testcae in self._testcases:
            if testcae.test_result:
                result=testcae.test_result
            else:
                break
        return result
    
    @property
    def test_resmgr(self):
        '''资源管理器
        '''
        return self._resmgr
    
    @test_resmgr.setter
    def test_resmgr(self, resmgr):
        self._resmgr = resmgr
        for it in self._testcases:
            it.test_resmgr = resmgr

    def dumps(self):
        '''序列化
        '''
        from testbase import serialization
        return [serialization.dumps(it) for it in self._testcases]
    
    def loads(self, buf ):
        '''反序列化
        '''
        from testbase import serialization
        self._testcases = [serialization.loads(it) for it in buf]

def debug_run_all():
    '''调试执行当前脚本的全部用例
    '''    
    from testbase.loader import TestLoader
    from testbase.runner import TestRunner
    from testbase.report import StreamTestReport
    tests = TestLoader().load("__main__")
    runner = TestRunner(StreamTestReport(output_testresult=True, output_summary=True))
    runner.run(tests)
    


#-*- coding: UTF-8 -*-
#
# Tencent is pleased to support the open source community by making QTA available.
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the BSD 3-Clause License (the "License"); you may not use this 
# file except in compliance with the License. You may obtain a copy of the License at
# 
# https://opensource.org/licenses/BSD-3-Clause
# 
# Unless required by applicable law or agreed to in writing, software distributed 
# under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#
"""Test plan
"""

from testbase.resource import TestResourceManager

class TestPlan(object):
    """测试计划
    """
    tests = None
    test_target_args = None

    def get_tests(self):
        """获取测试用例定义

        :rtypes: str/list(str)/list(TestCase)/TestCaseSettings
        """
        if self.tests is None:
            raise ValueError()
        return self.tests

    def get_test_target(self):
        """获取被测对象

        :return: 被测对象信息
        :rtypes: dict
        """
        if isinstance(self.test_target_args, dict):
            return self.test_target_args
        raise ValueError("`test_target_args` is not a dictionary, please overwrite `get_test_target` for your custom implenmentation")
        
    def test_setup(self, report):
        """测试初始化步骤
        """
        pass

    def test_teardown(self, report):
        """测试清理步骤
        """
        pass

    def resource_setup(self, report, restype, resource):
        """测试资源初始化
        """
        pass

    def resource_teardown(self, report, restype, resource):
        """测试资源清理
        """
        pass

    def debug_run(self, report=None, resmgr_backend=None):
        """调试执行
        """
        from testbase.runner import TestRunner
        from testbase.report import StreamTestReport
        if report is None:
            report = StreamTestReport()
        TestRunner(report, resmgr_backend).run(self)


# -*- coding: utf-8 -*-
#
# Tencent is pleased to support the open source community by making QTA available.
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the BSD 3-Clause License (the "License"); you may not use this 
# file except in compliance with the License. You may obtain a copy of the License at
# 
# https://opensource.org/licenses/BSD-3-Clause
# 
# Unless required by applicable law or agreed to in writing, software distributed 
# under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#
"""
打包
"""

import os
import sys
import re
import shutil
import zipfile
import pkg_resources
import subprocess

from testbase import resource
from testbase.conf import settings
from testbase.util import codecs_open

SETUP_PY_TEMPLATE = """
from setuptools import setup, find_packages

setup(
    version="%(version)s",
    name="%(name)s",
    packages=find_packages(),
    py_modules=["settings"],
    include_package_data = True,
    install_requires=%(requirements)s,
    entry_points={'console_scripts': ['qta-manage-venv = testbase.management:qta_manage_main'], },
)
"""


SETUP_PY_TEMPLATE_INCLUDE_RESOURCE = """
import os
import re
from setuptools import setup, find_packages, Command
    
BASE_DIR = os.path.dirname(os.path.abspath(__file__))

def find_all_file_patterns( resource_dir ):
    file_patterns = []
    rel_prefix = "." + os.path.sep
    for dirpath, dirnames, filenames in os.walk(resource_dir):
        if filenames:
            relpath = os.path.relpath(dirpath, resource_dir)
            pattern = os.path.join(relpath, "*")
            pattern = "resources" + os.path.sep + pattern.lstrip(rel_prefix)
            file_patterns.append(pattern)
    return file_patterns

def find_resource_packages_data():
    '''Collect files in 'resources' directory in package as package data
    '''
    packages_data = {}
    
    for dirpath, dirnames, _ in os.walk(BASE_DIR):
        for dirname in list(dirnames):
            if dirname == 'resources':
                if not os.path.isfile(os.path.join(dirpath, "__init__.py")):
                    continue
                resource_dir = os.path.join(dirpath, dirname)
                relpath = os.path.relpath(dirpath, BASE_DIR)
                package_name = ".".join(relpath.split(os.path.sep))
                packages_data[package_name] = find_all_file_patterns(resource_dir)

            if not os.path.isfile(os.path.join(dirpath, dirname, "__init__.py")):
                dirnames.remove(dirname)  # not a package
                continue
    return packages_data

def find_resource_data_files():
    '''Collect files in top level 'resources' directory as data_files
    '''
    import site
    if hasattr(site, "getsitepackages"):
        install_prefix = os.path.join(site.getsitepackages()[0], "resources")
    else:
        # Work around for virtuelenv bug: https://github.com/pypa/virtualenv/issues/355
        install_prefix = os.path.join(".", "lib", "site-packages", "resources")
    data_files = []
    resource_dir = os.path.join(BASE_DIR, "resources")
    for dirpath, dirnames, filenames in os.walk(resource_dir):
        dirrelpath = os.path.relpath(dirpath, resource_dir)
        filelist = [
            os.path.join("resources", dirrelpath, filename)
            for filename in filenames
        ]
        data_files.append((
            os.path.join(install_prefix, dirrelpath),
            filelist
        ))
    return data_files

class sdist_qta(Command):
    user_options = []
    boolean_options = []
    sub_commands = (("sdist_qta_resource", None),)
          
    def initialize_options (self):
        pass
    
    def finalize_options(self):
        pass

    def run(self):
        sdist = self.reinitialize_command("sdist")
        sdist.sub_commands = self.sub_commands
        self.run_command("sdist")
        
class sdist_qta_resource(Command):
    user_options = []
    boolean_options = []
        
    def initialize_options (self):
        pass
    
    def finalize_options(self):
        pass

    def run(self):
        sdist = self.get_finalized_command("sdist")
        src_root = os.getcwd()
        for dirpath, _, filenames in os.walk(src_root):
            relpath = os.path.relpath(dirpath, src_root)
            if "resources" in relpath.split(os.path.sep):
                for filename in filenames:
                    sdist.filelist.append(os.path.join(relpath, filename))
setup(
    version="%(version)s",
    name="%(name)s",
    # We want to bulid sdist with package datas without a MANIFEST.in
    # Work around for https://stackoverflow.com/questions/7522250/how-to-include-package-data-with-setuptools-distribute
    cmdclass={"sdist_qta": sdist_qta,
              "sdist_qta_resource": sdist_qta_resource},
    packages=find_packages(),
    py_modules=["settings"],
    package_data=find_resource_packages_data(),
    data_files=find_resource_data_files(),
    include_package_data = True,
    install_requires=%(requirements)s,
    entry_points={'console_scripts': ['qta-manage-venv = testbase.management:qta_manage_main'], },
)
"""


class DistGenerator(object):
    """Build dist packages"""

    def __init__(self, version):
        self._version = version

    def run(self, exclude_resources):
        self._generate_sdist(exclude_resources=exclude_resources)

    def _merge_requirements(self):
        """Merge exlib & requirements.txt"""
        exlib = os.path.join(settings.PROJECT_ROOT, 'exlib')
        egg_pattern = re.compile(r"(?P<name>[a-zA-Z0-9_]+)(\-(?P<version>[0-9a-zA-Z_\.]+)|)(\-.*|)\.egg$")
        reqs_dict = {}

        req_txt = os.path.join(settings.PROJECT_ROOT, "requirements.txt")
        if os.path.isfile(req_txt):
            with codecs_open(req_txt, 'r', encoding="utf-8") as fd:
                for it in pkg_resources.parse_requirements(fd.read()):
                    reqs_dict[it.name] = str(it)

        elif os.path.isdir(exlib):
            for filename in os.listdir(exlib):
                if not filename.endswith(".egg"):
                    continue
                result = egg_pattern.match(filename)
                if not result:
                    continue
                name, version = result.group('name'), result.group('version')
                if version:
                    reqs_dict[name] = "%s==%s" % (name, version)
                else:
                    reqs_dict[name] = name

        if 'qtaf' not in reqs_dict:
            reqs_dict["qtaf"] = "qtaf"
        return reqs_dict.values()

    def _generate_sdist(self, exclude_resources):
        """Call setuptools to generate source dist"""
        reqs = self._merge_requirements()
        setup_py = os.path.join(settings.PROJECT_ROOT, "setup.py")
        if exclude_resources:
            template = SETUP_PY_TEMPLATE
            cmd = "sdist"
        else:
            template = SETUP_PY_TEMPLATE_INCLUDE_RESOURCE
            cmd = "sdist_qta"

        with codecs_open(setup_py, "w", encoding="utf-8") as fd:
            fd.write(template % dict(
                version=self._version,
                name=settings.PROJECT_NAME,
                requirements=repr(reqs),
            ))

        subprocess.call(["python", "setup.py", cmd], cwd=settings.PROJECT_ROOT)
        os.remove(setup_py)

    def _generate_resource_pkg(self):
        filename = "%s-%s-resource.zip" % (settings.PROJECT_NAME, self._version)
        filepath = os.path.join(settings.PROJECT_ROOT, "dist", filename)
        with zipfile.ZipFile(filepath, "w") as zf:
            for res_path in resource.iter_resource_paths():
                for dirpath, _, filenames in os.walk(res_path):
                    for filename in filenames:
                        src = os.path.join(dirpath, filename)
                        zf.write(src, os.path.relpath(src, res_path))

class VirtuelEnv(object):
    """virtual env for QTA test project
    """

    VENV_ENV_NAME = "QTAF_VENV"

    def __init__(self, dist_pkg_path, path=None, recreate=False):
        self._dist_pkg_path = dist_pkg_path
        self._venv = path
        self._recreate_venv = recreate

    def activate(self):
        """activate virtuelenv on current processs

        Logic:
            1. Call from 'qta-manage', then create virtualenv
            2. Activate it
            3. Set OS environ 'QTAF_VENV'
            4. Create child process, replace 'qta-manage' with 'qta-manage-venv'
                (1) Call from 'qta-manage-venv'
                (2) OS environ 'QTAF_VENV' detected, ignore virtualenv creation code
        """
        
        if self.VENV_ENV_NAME in os.environ: # already in our venv
            return

        try:
            import virtualenv
        except ImportError:
            raise RuntimeError("Python package 'virtualenv' need to be not installed.")

        if self._venv:
            venv_path = self._venv
        else:
            venv_path = os.path.basename(self._dist_pkg_path)
            if venv_path.endswith(".zip"):
                venv_path = venv_path.rsplit(".", 1)[0]
            elif venv_path.endswith(".tar.gz"):
                venv_path = venv_path.rsplit(".", 2)[0]
            venv_path = venv_path + '_venv'
        if (not os.path.isdir(venv_path)) or self._recreate_venv:
            if os.path.isdir(venv_path):
                shutil.rmtree(venv_path)
            virtualenv.create_environment(venv_path, site_packages=True)
            created = True
        else:
            created = False
        _, _, _, bin_dir = virtualenv.path_locations(venv_path)
        activation_script = os.path.join(bin_dir, 'activate_this.py')
        execfile(activation_script, dict(__file__=activation_script))
        if created:
            subprocess.call(["pip", "install", self._dist_pkg_path], close_fds=True)

        os.environ[self.VENV_ENV_NAME] = "1"
        argv = list(sys.argv)
        argv[0] = "qta-manage-venv"
        subprocess.call(argv, close_fds=True)
        sys.exit(0)



# -*- coding: utf-8 -*-
#
# Tencent is pleased to support the open source community by making QTA available.
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the BSD 3-Clause License (the "License"); you may not use this 
# file except in compliance with the License. You may obtain a copy of the License at
# 
# https://opensource.org/licenses/BSD-3-Clause
# 
# Unless required by applicable law or agreed to in writing, software distributed 
# under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#
"""assert机制实现
"""

from __future__ import print_function

import ast
import copy
import inspect
import itertools
import pprint
import six
import sys
import threading
import traceback
import types

from testbase.util import Singleton, get_method_defined_class, smart_text

unary_map = {ast.Not: "not %s", ast.Invert: "~%s", ast.USub: "-%s", ast.UAdd: "+%s"}

binop_map = {
    ast.BitOr: "|",
    ast.BitXor: "^",
    ast.BitAnd: "&",
    ast.LShift: "<<",
    ast.RShift: ">>",
    ast.Add: "+",
    ast.Sub: "-",
    ast.Mult: "*",
    ast.Div: "/",
    ast.FloorDiv: "//",
    ast.Mod: "%%",  # escaped for string formatting
    ast.Eq: "==",
    ast.NotEq: "!=",
    ast.Lt: "<",
    ast.LtE: "<=",
    ast.Gt: ">",
    ast.GtE: ">=",
    ast.Pow: "**",
    ast.Is: "is",
    ast.IsNot: "is not",
    ast.In: "in",
    ast.NotIn: "not in",
}

if sys.version_info >= (3, 5):
    ast_Call = ast.Call
else:

    def ast_Call(a, b, c):
        return ast.Call(a, b, c, None, None)
    
if hasattr(ast, "NameConstant"):
    _NameConstant = ast.NameConstant
else:

    def _NameConstant(c):
        return ast.Name(str(c), ast.Load())
    
def set_location(node, lineno, col_offset):
    """Set node location information recursively."""

    def _fix(node, lineno, col_offset):
        if "lineno" in node._attributes:
            node.lineno = lineno
        if "col_offset" in node._attributes:
            node.col_offset = col_offset
        for child in ast.iter_child_nodes(node):
            _fix(child, lineno, col_offset)

    _fix(node, lineno, col_offset)
    return node

def hook_function(func):
    code = func.__code__
    src_code = inspect.getsource(code)
    tree = ast.parse(src_code, "<string>", "exec")
    print_func = ast.Name(id="print",ctx=ast.Load())
    arg = ast.Str("hello world!")
    print_expr = ast.Expr(ast.Call(func=print_func, args=[arg], keywords=[]))
    test_func = tree.body[0]
    test_func.body.append(print_expr)
    ast.fix_missing_locations(tree)
    new_code = compile(tree, "<string>", "exec", dont_inherit=True)
    for item in new_code.co_consts:
        if isinstance(item, types.CodeType) and item.co_name == code.co_name:
            func.__code__ = copy.deepcopy(item)
            break
    else:
        raise RuntimeError("no match name function for %s" % code.co_name)
    
class AssertionRewriter(ast.NodeVisitor):
    """assert rewriter
    """
    
    def rewrite(self, item):
        try:
            self.rewrite_(item)
        except:
            stack = traceback.format_exc()
            msg = "[WARN]rewrite item %s failed: %s" % (item, stack)
            print(msg, file=sys.stderr)

    def rewrite_(self, item):
        """Find all assert statements in *mod* and rewrite them."""
        func = item
        if func in _AssertHookedCache():
            return
        
        mod, func_node = get_func_mod_and_node(func)
        if mod is None and func_node is None:
            _AssertHookedCache().add(func)
            return
        
        #compatibility with py 2 and 3
        if sys.version_info[0] == 3:
            builtins_mod = "builtins"
        else:
            builtins_mod = "__builtin__"
        aliases = [
            ast.alias(builtins_mod, "_py_builtins_"),
            ast.alias("testbase.assertion", "_qtaf_assert_"),
        ]
        pos = 0
        lineno = func_node.lineno
        col_offset = func_node.col_offset 
        
        imports = [
            ast.Import([alias], lineno=lineno, col_offset=col_offset) for alias in aliases
        ]
        imports.append(ast.ImportFrom(module="testbase.testresult", names=[ast.alias('EnumLogLevel',None)], level=0, lineno=lineno, col_offset=col_offset))
        func_node.body[pos:pos] = imports
        nodes = [func_node]
        self.rewrite_code = False
        while nodes:
            node = nodes.pop()
            for field, value in ast.iter_fields(node):
                if isinstance(value, list):
                    new_nodes = []
                    for child in value:
                        if isinstance(child, ast.Expr):
                            new_nodes.extend(self.visit(child))
                        else:
                            new_nodes.append(child)
                    setattr(node, field, new_nodes)
                if isinstance(value, ast.Expr):
                    self.visit(value)
        if self.rewrite_code:
            source_file = inspect.getsourcefile(func)
            new_code = compile(mod, source_file, "exec")
            set_func_code(func, new_code)
        _AssertHookedCache().add(func)
    
    def visit_Expr(self, expr):
        value = expr.value
        if isinstance(value, ast.Call):
            if isinstance(value.func, ast.Attribute):
                if value.func.attr == "assert_":
                    return self.rewrite_assert_(expr)
            elif isinstance(value.func, ast.Name):
                if value.func.id == "assert_":
                    return self.rewrite_assert_(value)
            
        return [expr]
            
    def rewrite_assert_(self, elem):
        """rewrite the assert_ expression
        """
        self.rewrite_code = True
        self.statements = []
        self.variables = []
        self.variable_counter = itertools.count()
        self.stack = []
        self.on_failure = []
        self.push_format_context()
        
        # Rewrite assert into a bunch of statements.
        if isinstance(elem, ast.Expr):
            top_condition, explanation = self.visit(elem.value.args[1])
            msg_arg = elem.value.args[0]
            caller_id = "self"
        elif isinstance(elem, ast.Call):
            top_condition, explanation = self.visit(elem.args[1])
            msg_arg = elem.args[0]
            caller_id = "_qtaf_assert_"

        # Create failure message.
        body = self.on_failure
        negation = ast.UnaryOp(ast.Not(), top_condition)
        self.statements.append(ast.If(negation, body, []))
        assertmsg = self.helper("format_assertmsg", msg_arg)
        explanation = " " + explanation
        template = ast.BinOp(assertmsg, ast.Add(), ast.Str(explanation))
        msg = self.pop_format_context(template)
        fmt = self.helper("format_explanation", msg)
        log_record = ast.Attribute(
                                value=ast.Name(id=caller_id, ctx=ast.Load()), 
                                attr='_log_assert_failed', 
                                ctx=ast.Load())
        args = [fmt]
        exc = ast_Call(log_record, args, [])
        log_record_expr = ast.Expr(value=exc)
        body.append(log_record_expr)
        # Clear temporary variables by setting them to None.
        if self.variables:
            variables = [ast.Name(name, ast.Store()) for name in self.variables]
            clear = ast.Assign(variables, _NameConstant(None))
            self.statements.append(clear)
        # Fix line numbers.
        for stmt in self.statements:
            set_location(stmt, elem.lineno, elem.col_offset)
        return self.statements
        
    def variable(self):
        """Get a new variable."""
        # Use a character invalid in python identifiers to avoid clashing.
        name = "_py_assert" + str(next(self.variable_counter))
        self.variables.append(name)
        return name

    def assign(self, expr):
        """Give *expr* a name."""
        name = self.variable()
        self.statements.append(ast.Assign([ast.Name(name, ast.Store())], expr))
        return ast.Name(name, ast.Load())

    def display(self, expr):
        """Call py.io.saferepr on the expression."""
        return self.helper("saferepr", expr)

    def helper(self, name, *args):
        """Call a helper in this module."""
        py_name = ast.Name("_qtaf_assert_", ast.Load())
        attr = ast.Attribute(py_name, "_" + name, ast.Load())
        return ast_Call(attr, list(args), [])

    def builtin(self, name):
        """Return the builtin called *name*."""
        builtin_name = ast.Name("_py_builtins_", ast.Load())
        return ast.Attribute(builtin_name, name, ast.Load())

    def explanation_param(self, expr):
        """Return a new named %-formatting placeholder for expr.

        This creates a %-formatting placeholder for expr in the
        current formatting context, e.g. ``%(py0)s``.  The placeholder
        and expr are placed in the current format context so that it
        can be used on the next call to .pop_format_context().

        """
        specifier = "py" + str(next(self.variable_counter))
        self.explanation_specifiers[specifier] = expr
        return "%(" + specifier + ")s"

    def push_format_context(self):
        """Create a new formatting context.

        The format context is used for when an explanation wants to
        have a variable value formatted in the assertion message.  In
        this case the value required can be added using
        .explanation_param().  Finally .pop_format_context() is used
        to format a string of %-formatted values as added by
        .explanation_param().

        """
        self.explanation_specifiers = {}
        self.stack.append(self.explanation_specifiers)

    def pop_format_context(self, expl_expr):
        """Format the %-formatted string with current format context.

        The expl_expr should be an ast.Str instance constructed from
        the %-placeholders created by .explanation_param().  This will
        add the required code to format said string to .on_failure and
        return the ast.Name instance of the formatted string.

        """
        current = self.stack.pop()
        if self.stack:
            self.explanation_specifiers = self.stack[-1]
        keys = [ast.Str(key) for key in current.keys()]
        format_dict = ast.Dict(keys, list(current.values()))
        form = ast.BinOp(expl_expr, ast.Mod(), format_dict)
        name = "_py_format_" + str(next(self.variable_counter))
        self.on_failure.append(ast.Assign([ast.Name(name, ast.Store())], form))
        return ast.Name(name, ast.Load())

    def generic_visit(self, node):
        """Handle expressions we don't have custom code for."""
        assert isinstance(node, ast.expr)
        res = self.assign(node)
        return res, self.explanation_param(self.display(res))

    def visit_Name(self, name):
        # Display the repr of the name if it's a local variable or
        # _should_repr_global_name() thinks it's acceptable.
        locs = ast_Call(self.builtin("locals"), [], [])
        inlocs = ast.Compare(ast.Str(name.id), [ast.In()], [locs])
        dorepr = self.helper("should_repr_global_name", name)
        test = ast.BoolOp(ast.Or(), [inlocs, dorepr])
        expr = ast.IfExp(test, self.display(name), ast.Str(name.id))
        return name, self.explanation_param(expr)

    def visit_BoolOp(self, boolop):
        res_var = self.variable()
        expl_list = self.assign(ast.List([], ast.Load()))
        app = ast.Attribute(expl_list, "append", ast.Load())
        is_or = int(isinstance(boolop.op, ast.Or))
        body = save = self.statements
        fail_save = self.on_failure
        levels = len(boolop.values) - 1
        self.push_format_context()
        # Process each operand, short-circuting if needed.
        cond = None
        for i, v in enumerate(boolop.values):
            if i:
                fail_inner = []
                # cond is set in a prior loop iteration below
                self.on_failure.append(ast.If(cond, fail_inner, []))  # noqa
                self.on_failure = fail_inner
            self.push_format_context()
            res, expl = self.visit(v)
            body.append(ast.Assign([ast.Name(res_var, ast.Store())], res))
            expl_format = self.pop_format_context(ast.Str(expl))
            call = ast_Call(app, [expl_format], [])
            self.on_failure.append(ast.Expr(call))
            if i < levels:
                cond = res
                if is_or:
                    cond = ast.UnaryOp(ast.Not(), cond)
                inner = []
                self.statements.append(ast.If(cond, inner, []))
                self.statements = body = inner
        self.statements = save
        self.on_failure = fail_save
        expl_template = self.helper("format_boolop", expl_list, ast.Num(is_or))
        expl = self.pop_format_context(expl_template)
        return ast.Name(res_var, ast.Load()), self.explanation_param(expl)

    def visit_UnaryOp(self, unary):
        pattern = unary_map[unary.op.__class__]
        operand_res, operand_expl = self.visit(unary.operand)
        res = self.assign(ast.UnaryOp(unary.op, operand_res))
        return res, pattern % (operand_expl,)

    def visit_BinOp(self, binop):
        symbol = binop_map[binop.op.__class__]
        left_expr, left_expl = self.visit(binop.left)
        right_expr, right_expl = self.visit(binop.right)
        explanation = "(%s %s %s)" % (left_expl, symbol, right_expl)
        res = self.assign(ast.BinOp(left_expr, binop.op, right_expr))
        return res, explanation

    def visit_Call_35(self, call):
        """
        visit `ast.Call` nodes on Python3.5 and after
        """
        new_func, func_expl = self.visit(call.func)
        arg_expls = []
        new_args = []
        new_kwargs = []
        for arg in call.args:
            res, expl = self.visit(arg)
            arg_expls.append(expl)
            new_args.append(res)
        for keyword in call.keywords:
            res, expl = self.visit(keyword.value)
            new_kwargs.append(ast.keyword(keyword.arg, res))
            if keyword.arg:
                arg_expls.append(keyword.arg + "=" + expl)
            else:  # **args have `arg` keywords with an .arg of None
                arg_expls.append("**" + expl)

        expl = "%s(%s)" % (func_expl, ", ".join(arg_expls))
        new_call = ast.Call(new_func, new_args, new_kwargs)
        res = self.assign(new_call)
        res_expl = self.explanation_param(self.display(res))
        outer_expl = "%s\n{%s = %s\n}" % (res_expl, res_expl, expl)
        return res, outer_expl

    def visit_Starred(self, starred):
        # From Python 3.5, a Starred node can appear in a function call
        _, expl = self.visit(starred.value)
        return starred, "*" + expl

    def visit_Call_legacy(self, call):
        """
        visit `ast.Call nodes on 3.4 and below`
        """
        new_func, func_expl = self.visit(call.func)
        arg_expls = []
        new_args = []
        new_kwargs = []
        new_star = new_kwarg = None
        for arg in call.args:
            res, expl = self.visit(arg)
            new_args.append(res)
            arg_expls.append(expl)
        for keyword in call.keywords:
            res, expl = self.visit(keyword.value)
            new_kwargs.append(ast.keyword(keyword.arg, res))
            arg_expls.append(keyword.arg + "=" + expl)
        if call.starargs:
            new_star, expl = self.visit(call.starargs)
            arg_expls.append("*" + expl)
        if call.kwargs:
            new_kwarg, expl = self.visit(call.kwargs)
            arg_expls.append("**" + expl)
        expl = "%s(%s)" % (func_expl, ", ".join(arg_expls))
        new_call = ast.Call(new_func, new_args, new_kwargs, new_star, new_kwarg)
        res = self.assign(new_call)
        res_expl = self.explanation_param(self.display(res))
        outer_expl = "%s\n{%s = %s\n}" % (res_expl, res_expl, expl)
        return res, outer_expl

    # ast.Call signature changed on 3.5,
    # conditionally change  which methods is named
    # visit_Call depending on Python version
    if sys.version_info >= (3, 5):
        visit_Call = visit_Call_35
    else:
        visit_Call = visit_Call_legacy

    def visit_Attribute(self, attr):
        if not isinstance(attr.ctx, ast.Load):
            return self.generic_visit(attr)
        value, value_expl = self.visit(attr.value)
        res = self.assign(ast.Attribute(value, attr.attr, ast.Load()))
        res_expl = self.explanation_param(self.display(res))
        pat = "%s\n{%s = %s.%s\n}"
        expl = pat % (res_expl, res_expl, value_expl, attr.attr)
        return res, expl

    def visit_Compare(self, comp):
        self.push_format_context()
        left_res, left_expl = self.visit(comp.left)
        if isinstance(comp.left, (ast.Compare, ast.BoolOp)):
            left_expl = "({})".format(left_expl)
        res_variables = [self.variable() for i in range(len(comp.ops))]
        load_names = [ast.Name(v, ast.Load()) for v in res_variables]
        store_names = [ast.Name(v, ast.Store()) for v in res_variables]
        it = zip(range(len(comp.ops)), comp.ops, comp.comparators)
        expls = []
        syms = []
        results = [left_res]
        for i, op, next_operand in it:
            next_res, next_expl = self.visit(next_operand)
            if isinstance(next_operand, (ast.Compare, ast.BoolOp)):
                next_expl = "({})".format(next_expl)
            results.append(next_res)
            sym = binop_map[op.__class__]
            syms.append(ast.Str(sym))
            expl = "%s %s %s" % (left_expl, sym, next_expl)
            expls.append(ast.Str(expl))
            res_expr = ast.Compare(left_res, [op], [next_res])
            self.statements.append(ast.Assign([store_names[i]], res_expr))
            left_res, left_expl = next_res, next_expl
        # Use pytest.assertion.util._reprcompare if that's available.
        expl_call = self.helper(
            "call_reprcompare",
            ast.Tuple(syms, ast.Load()),
            ast.Tuple(load_names, ast.Load()),
            ast.Tuple(expls, ast.Load()),
            ast.Tuple(results, ast.Load()),
        )
        if len(comp.ops) > 1:
            res = ast.BoolOp(ast.And(), load_names)
        else:
            res = load_names[0]
        return res, self.explanation_param(self.pop_format_context(expl_call))

def _call_reprcompare(ops, results, expls, each_obj):
    for _, res, expl in zip(range(len(ops)), results, expls):
        try:
            done = not res
        except Exception:
            done = True
        if done:
            break
    return expl

def _saferepr(obj):
    """Get a safe repr of an object for assertion error messages.

    The assertion formatting (util.format_explanation()) requires
    newlines to be escaped since they are a special character for it.
    Normally assertion.util.format_explanation() does this but for a
    custom repr it is possible to contain one of the special escape
    sequences, especially '\n{' and '\n}' are likely to be present in
    JSON reprs.

    """
    if isinstance(obj, types.MethodType):
        if obj.__self__:
            obj_repr = "self.%s" % obj.__func__.__name__
    elif isinstance(obj, type):
        obj_repr = obj.__name__
    elif isinstance(obj, six.string_types):
        obj_repr = '"%s"' % smart_text(obj)
    else:
        obj_repr = pprint.saferepr(obj)
    return obj_repr

def _should_repr_global_name(obj):
    return not hasattr(obj, "__name__") and not callable(obj)

def _format_assertmsg(obj):
    """Format the custom assertion message given.

    For strings this simply replaces newlines with '\n~' so that
    util.format_explanation() will preserve them instead of escaping
    newlines.  For other objects py.io.saferepr() is used first.

    """
    # reprlib appears to have a bug which means that if a string
    # contains a newline it gets escaped, however if an object has a
    # .__repr__() which contains newlines it does not get escaped.
    # However in either case we want to preserve the newline.
#     if isinstance(obj, six.text_type) or isinstance(obj, six.binary_type):
    if isinstance(obj, six.string_types):
        s = smart_text(obj)
        is_repr = False
    else:
        s = pprint.saferepr(obj)
        is_repr = True
    t = str
    s = s.replace(t("\n"), t("\n~")).replace(t("%"), t("%%"))
    if is_repr:
        s = s.replace(t("\\n"), t("\n~"))
    return s + "\n"

def _format_explanation(explanation):
    """This formats an explanation

    Normally all embedded newlines are escaped, however there are
    three exceptions: \n{, \n} and \n~.  The first two are intended
    cover nested explanations, see function and attribute explanations
    for examples (.visit_Call(), visit_Attribute()).  The last one is
    for when one explanation needs to span multiple lines, e.g. when
    displaying diffs.
    """
    raw_lines = (explanation or '').split('\n')
    # escape newlines not followed by {, } and ~
    msg = raw_lines[0]
    raw_lines = raw_lines[1:]
    lines = raw_lines[:1]
    for l in raw_lines[1:]:
        if l.startswith('{') or l.startswith('}') or l.startswith('~'):
            lines.append(l)
        else:
            lines[-1] += '\\n' + l

    result = lines[:1]
    stack = [0]
    stackcnt = [0]
    for line in lines[1:]:
        if line.startswith('{'):
            if stackcnt[-1]:
                s = 'and   '
            else:
                s = 'where '
            stack.append(len(result))
            stackcnt[-1] += 1
            stackcnt.append(0)
            result.append(' +' + '  '*(len(stack)-1) + s + line[1:])
        elif line.startswith('}'):
            stack.pop()
            stackcnt.pop()
            result[stack[-1]] += line[1:]
        else:
            assert line.startswith('~')
            result.append('  '*len(stack) + line[1:])
    assert len(stack) == 1
    result[0] = " [%s] assert " % msg + result[0]
    return '\n'.join(result)

def get_func_name(func):
    if isinstance(func, types.MethodType):
        if sys.version_info[0] == 3:
            return func.__func__.__name__
        else:
            return func.__name__
    else:
        return func.__name__

def get_func_source_code(func):
    if isinstance(func, types.MethodType):
        if sys.version_info[0] == 3:
            src_code = inspect.getsource(func.__self__.__class__)
        else:
            im_class = get_method_defined_class(func)
            src_code = inspect.getsource(im_class)
    else:
        src_code = inspect.getsource(func)
    indent = inspect.indentsize(src_code)
    if indent > 0:
        lines = src_code.split("\n")
        src_code = ""
        for line in lines:
            src_code += line[indent:] + "\n"
    return src_code 

def get_func_mod_and_node(func):
    func_name = get_func_name(func)
    src_code = get_func_source_code(func)
    if isinstance(func, types.MethodType):
        mod = ast.parse(src_code)
        ast_cls = mod.body[0]
        for item in ast_cls.body:
            if isinstance(item, ast.FunctionDef):
                if item.name == func_name:
                    func_node = item
                    break
        else:
            return None, None
    else:
        mod = ast.parse(src_code)
        func_node = mod.body[0]
    ast.increment_lineno(func_node, func.__code__.co_firstlineno - func_node.lineno)
    return mod, func_node

def get_func_compiled_code(func, new_code):
    func_name = get_func_name(func)
    if isinstance(func, types.MethodType):
        for item in new_code.co_consts:
            if isinstance(item, types.CodeType):
                for sub_item in item.co_consts:
                    if isinstance(sub_item, types.CodeType) and sub_item.co_name == func_name:
                        return sub_item
    elif isinstance(func, types.FunctionType):
        for item in new_code.co_consts:
            if isinstance(item, types.CodeType) and item.co_name == func_name:
                return item
    else:
        raise RuntimeError("%s not supported yet" % repr(func))
    
def set_func_code(func, new_code):
    new_func_code = get_func_compiled_code(func, new_code)
    if isinstance(func, types.MethodType):
        if sys.version_info[0] == 3:
            func.__code__ = new_func_code
        else:
            func.__func__.__code__ = new_func_code
    else:
        func.__code__ = new_func_code
    
class _AssertHookedCache(six.with_metaclass(Singleton, object)):
    
    def __init__(self):
        self._lock = threading.Lock()
        self.__cache = set()
        
    def add(self, func):
        with self._lock:
            self.__cache.add(self._hash_func(func))
            
    def __contains__(self, func):
        item =self._hash_func(func)
        return item in self.__cache
    
    def __iter__(self):
        return self.__cache.__iter__()
    
    def _hash_func(self, func):
        if isinstance(func, (types.FunctionType, types.MethodType)):
            return func
        else:
            raise ValueError("func must be a callable type or object")
        
if __name__ == "__main__":
    pass


# -*- coding: utf-8 -*-
#
# Tencent is pleased to support the open source community by making QTA available.
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the BSD 3-Clause License (the "License"); you may not use this 
# file except in compliance with the License. You may obtain a copy of the License at
# 
# https://opensource.org/licenses/BSD-3-Clause
# 
# Unless required by applicable law or agreed to in writing, software distributed 
# under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#
'''数据驱动模块

使用介绍：
1、在TestCase类前面加入@DataDrive装饰器，参数为测试数据列表或者字典::

    @datadrive.DataDrive([1,2,3])
    class Test(tc.TestCase):
        def runTest(self):
            pass
    
    @datadrive.DataDrive(
        {
        'TEST1':1,
        'TEST2':2,
        'TEST3':3,
        }
        )
    class Test(tc.TestCase):
        def runTest(self):
            pass
        
在测试报告中展示用例名字时，若参数为列表的，则在用例后面增加一个索引序号，若参数为字典的，则在用例后面增加上key的名字：

    - 列表显示为::
    
        Test#1
        Test#2
        Test#3
        
    - 字典显示为::

        Test#TEST1
        Test#TEST2
        Test#TEST3

2、可以在runTest里通过self.casedata使用测试数据::

    def runTest(self):
        print(self.casedata)
    
3、运行及调试方法和原来一样::

    MyTest().run()
        

    完整的例子如下::
    
    # -*- coding: utf-8 -*-
    
    import testbase.testcase as tc
    import testbase.datadrive as datadrive
    
    @datadrive.DataDrive([1,2,3])
    class MyTest(tc.TestCase):
        """test
        """
        
        owner = 'foo'
        priority = tc.TestCase.EnumPriority.High
        status = tc.TestCase.EnumStatus.Ready
        timeout = 5       
        
        def runTest(self):
            print 'runTest, casedata:', self.casedata
              
            
    if __name__ == '__main__':
        MyTest().run()
        
'''

import types
from testbase.testcase import TestCase

class DataDrive(object):
    '''数据驱动类修饰器，标识一个测试用例类使用数据驱动
    '''    
    def __init__(self, case_datas ):
        '''构造函数
        
        :param case_datas: 数据驱动测试数据集
        :type case_datas: list/tuple/dict
        '''
        self._case_datas = case_datas
                
    def __call__(self, testcase_class ):
        '''修饰器
        
        :param testcase_class: 要修饰的测试用例
        :type testcase_class: TestCase
        '''
        if not issubclass(testcase_class, TestCase):
            raise TypeError('数据驱动使用错误，只能对testcase类使用')
        testcase_class.__qtaf_datadrive__ = self
        return testcase_class
    
    def __iter__(self):
        '''遍历全部的数据名
        '''
        if isinstance(self._case_datas, types.GeneratorType):
            self._case_datas = list(self._case_datas)
        if isinstance(self._case_datas, list) or isinstance(self._case_datas, tuple):
            for it in range(len(self._case_datas)):
                yield it
        else:
            for it in self._case_datas:
                yield it
                
    def __getitem__(self, name):
        '''获取对应名称的数据
        '''
        return self._case_datas[name]
    
    def __len__(self):
        return len(self._case_datas)
        
def is_datadrive( obj ):
    '''是否为数据驱动用例
    
    :param obj: 测试用例或测试用例类
    :type obj: TestCase/type
    :returns boolean
    '''
    return hasattr(obj, '__qtaf_datadrive__')

def get_datadrive( obj ):
    '''获取对应用例的数据驱动
    
    :param obj: 测试用例或测试用例类
    :type obj: TestCase/type
    :returns DataDrive
    '''
    return obj.__qtaf_datadrive__

def load_datadrive_tests( cls, name=None ):
    '''加载对应数据驱动测试用例类的数据驱动用例
    '''
    dd = get_datadrive(cls)
    if name:
        if name not in dd:
            raise ValueError("找不到对应名字'%s'的数据驱动用例"%name)
        testdata = dd[name]
        if isinstance(testdata, dict) and "__attrs__" in testdata:
            attrs = testdata["__attrs__"]
        else:
            attrs = None
        tests = [cls(dd[name], str(name), attrs)]
    else:
        tests = []
        for name in dd:
            testdata = dd[name]
            if isinstance(testdata, dict) and "__attrs__" in testdata:
                attrs = testdata["__attrs__"]
            else:
                attrs = None
            tests.append(cls(dd[name], str(name), attrs))
    return tests



# -*- coding: utf-8 -*-
#
# Tencent is pleased to support the open source community by making QTA available.
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the BSD 3-Clause License (the "License"); you may not use this 
# file except in compliance with the License. You may obtain a copy of the License at
# 
# https://opensource.org/licenses/BSD-3-Clause
# 
# Unless required by applicable law or agreed to in writing, software distributed 
# under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#
'''测试报告
'''

import sys
import socket
import os
import shutil
import json
import getpass
import locale
import argparse
import pkg_resources
import xml.dom.minidom as dom
import xml.sax.saxutils as saxutils
import traceback

from datetime import datetime

from testbase import logger
from testbase import testresult
from testbase.testresult import EnumLogLevel
from testbase.util import smart_text, smart_binary, to_pretty_xml, ensure_binary_stream, \
    codecs_open, get_os_version, get_inner_resource
    
REPORT_ENTRY_POINT = "qtaf.report"
report_types = {}
os_encoding = locale.getdefaultlocale()[1]
report_usage = 'runtest <test ...> --report-type <report-type> [--report-args "<report-args>"]'
        
class ITestReport(object):
    '''测试报告接口
    '''
    def begin_report(self):
        '''开始测试执行
        '''
        pass
    
    def end_report(self):
        '''结束测试执行

        :param passed: 测试是否通过
        :type passed: boolean
        '''
        pass
        
    def log_test_result(self, testcase, testresult ):
        '''记录一个测试结果

        :param testcase: 测试用例
        :type testcase: TestCase
        :param testresult: 测试结果
        :type testresult: TestResult
        '''
        pass
    
    def log_record(self, level, tag, msg, record):
        '''增加一个记录

        :param level: 日志级别
        :param msg: 日志消息
        :param tag: 日志标签
        :param record: 日志记录信息
        :type level: string
        :type tag: string
        :type msg: string
        :type record: dict
        '''
        pass

    def log_loaded_tests(self, loader, testcases):
        '''记录加载成功的用例

        :param loader: 用例加载器
        :type loader: TestLoader
        :param testcases: 测试用例列表
        :type testcases: list
        '''
        pass

    def log_filtered_test(self, loader, testcase, reason):
        '''记录一个被过滤的测试用例

        :param loader: 用例加载器
        :type loader: TestLoader
        :param testcase: 测试用例
        :type testcase: TestCase
        :param reason: 过滤原因
        :type reason: str
        '''
        pass

    def log_load_error(self, loader, name, error):
        '''记录一个加载失败的用例或用例集

        :param loader: 用例加载器
        :type loader: TestLoader
        :param name: 名称
        :type name: str
        :param error: 错误信息
        :type error: str
        '''
        pass

    def log_test_target(self, test_target):
        '''记录被测对象

        :param test_target: 被测对象详情
        :type test_target: any
        '''
        pass

    def log_resource(self, res_type, resource):
        '''记录测试使用的资源

        :param res_type: 资源类型
        :type res_type: str
        :param resource: 资源详情
        :type resource: dict
        '''
        pass
    
    def get_testresult_factory(self):
        '''获取对应的TestResult工厂

        :returns ITestResultFactory
        '''
        raise NotImplementedError()
    
    def debug(self, tag, msg, record=None):
        '''记录一个DEBUG日志
        :param msg: 日志消息
        :param tag: 日志标签
        :param record: 日志记录信息
        :type tag: string
        :type msg: string
        :type record: dict
        '''
        if record is None:
            record = {}
        self.log_record(EnumLogLevel.DEBUG, tag, msg, record)
        
    def info(self, tag, msg, record=None):
        '''记录一个INFO日志
        :param msg: 日志消息
        :param tag: 日志标签
        :param record: 日志记录信息
        :type tag: string
        :type msg: string
        :type record: dict
        '''
        if record is None:
            record = {}
        self.log_record(EnumLogLevel.INFO, tag, msg, record)
        
    def warning(self, tag, msg, record=None):
        '''记录一个WARN日志
        :param msg: 日志消息
        :param tag: 日志标签
        :param record: 日志记录信息
        :type tag: string
        :type msg: string
        :type record: dict
        '''
        if record is None:
            record = {}
        self.log_record(EnumLogLevel.WARNING, tag, msg, record)
        
    def error(self, tag, msg, record=None):
        '''记录一个ERROR日志
        :param msg: 日志消息
        :param tag: 日志标签
        :param record: 日志记录信息
        :type tag: string
        :type msg: string
        :type record: dict
        '''
        if record is None:
            record = {}
        self.log_record(EnumLogLevel.ERROR, tag, msg, record)
        
    def critical(self, tag, msg, record=None):
        '''记录一个CRITICAL日志
        :param msg: 日志消息
        :param tag: 日志标签
        :param record: 日志记录信息
        :type tag: string
        :type msg: string
        :type record: dict
        '''
        if record is None:
            record = {}
        self.log_record(EnumLogLevel.CRITICAL, tag, msg, record)

    @classmethod
    def get_parser(cls):
        '''获取命令行参数解析器（如果实现）

        :returns: 解析器对象
        :rtype: argparse.ArgumentParser
        '''
        raise NotImplementedError()

    @classmethod
    def parse_args(cls, args_string):
        '''通过命令行参数构造对象
        
        :returns: 测试报告
        :rtype: cls
        '''
        raise NotImplementedError()
        
class ITestResultFactory(object):
    '''TestResult工厂接口
    '''
    def create(self, testcase ):
        '''创建TestResult对象
        :param testcase: 测试用例
        :type testcase: TestCase
        :return TestResult
        '''
        raise NotImplementedError()
    
    def dumps(self):
        '''序列化
        :return picklable object
        '''
        pass
    
    def loads(self, buf):
        '''反序列化
        :param buf: dumps返回的序列化后的数据
        :type buf: object
        '''
        pass
        
class EmptyTestResultFactory(ITestResultFactory):
    '''测试结果工厂
    '''
    def __init__(self, result_factory_func=None ):
        '''构造函数
        :param result_factory_func: TestResult工厂函数
        :type result_factory_func: Function
        '''
        self._result_factory_func = result_factory_func
        
    def create(self, testcase ):
        '''创建TestResult对象
        :param testcase: 测试用例
        :type testcase: TestCase
        :return TestResult
        '''
        if self._result_factory_func is None:
            return testresult.EmptyResult()
        else:
            return self._result_factory_func(testcase)
        
    def dumps(self):
        '''序列化
        :return picklable object
        '''
        return self._result_factory_func
    
    def loads(self, buf):
        '''反序列化
        :param buf: dumps返回的序列化后的数据
        :type buf: object
        '''
        self._result_factory_func = buf
        
class EmptyTestReport(ITestReport):
    '''不输出测试报告
    '''
    def __init__(self, result_factory_func=None ):
        '''构造函数
        :param result_factory_func: TestResult工厂函数
        :type result_factory_func: callable
        '''
        self._result_factory_func = result_factory_func
        self._is_passed = True
    
    def get_testresult_factory(self):
        '''获取对应的TestResult工厂
        :returns ITestResultFactory
        '''
        return EmptyTestResultFactory(self._result_factory_func)

    def log_test_result(self, testcase, testresult ):
        '''记录一个测试结果

        :param testcase: 测试用例
        :type testcase: TestCase
        :param testresult: 测试结果
        :type testresult: TestResult
        '''
        if not testresult.passed:
            self._is_passed = False
    
    @property
    def passed(self):
        '''测试是否通过
        '''
        return self._is_passed

    @classmethod
    def get_parser(cls):
        '''获取命令行参数解析器（如果实现）

        :returns: 解析器对象
        :rtype: argparse.ArgumentParser
        '''
        return argparse.ArgumentParser(usage=report_usage)

    @classmethod
    def parse_args(cls, args_string):
        '''通过命令行参数构造对象
        
        :returns: 测试报告
        :rtype: cls
        '''
        return EmptyTestReport()

class StreamTestResultFactory(ITestResultFactory):
    '''流形式TestResult工厂
    '''
    def __init__(self, stream ):
        '''构造函数
        :param stream: 指定要输出的流设备
        :type stream: file
        '''
        self._stream = stream
        
    def create(self, testcase ):
        '''创建TestResult对象
        :param testcase: 测试用例
        :type testcase: TestCase
        :return TestResult
        '''
        return testresult.StreamResult(self._stream)
    
    def dumps(self):
        '''序列化
        :return picklable object
        '''
        fileno = self._stream.fileno()
        if fileno not in [0, 1]:
            raise ValueError("不支持的流对象: %s" % self._stream)
        return fileno
    
    def loads(self, buf):
        '''反序列化
        :param buf: dumps返回的序列化后的数据
        :type buf: object
        '''
        fileno = buf
        if fileno == 1:
            self._stream = sys.stdout
        elif fileno == 2:
            self._stream = sys.stderr
        else:
            raise ValueError("invalid fd: %s" % fileno )
    
class StreamTestReport(ITestReport):
    '''流形式的测试报告
    '''
    def __init__(self, stream=sys.stdout, error_stream=sys.stderr, output_testresult=False, output_summary=True ):
        '''构造函数
        :param stream: 指定要输出的流设备
        :type stream: file
        :param output_testresult: 是否输出测试用例执行的日志
        :type output_testresult: boolean
        :param output_summary: 是否输出执行汇总信息
        :type output_summary: boolean
        '''
        self._stream, encoding = ensure_binary_stream(stream)
        self._err_stream, _ = ensure_binary_stream(error_stream)
        self._write = lambda x: self._stream.write(smart_binary(x, encoding=encoding))
        self._write_err = lambda x: self._err_stream.write(smart_binary(x, encoding=encoding))
                    
        self._output_testresult = output_testresult
        self._output_summary = output_summary
        self._passed_testresults = []
        self._failed_testresults = [] 
            
    def begin_report(self):
        '''开始测试执行
        '''
        self._start_time = datetime.now()
        self._write("Test runs at:%s.\n" % self._start_time.strftime("%Y-%m-%d %H:%M:%S"))
    
    def end_report(self):
        '''结束测试执行
        :param passed: 测试是否通过
        :type passed: boolean
        '''
        end_time = datetime.now()
        self._write("Test ends at:%s.\n" % end_time.strftime("%Y-%m-%d %H:%M:%S"))
        #self._write("Total execution time is :%s\n" % str(end_time-self._start_time).split('.')[0])
    
        if self._output_summary:
            self._write("\n" + "="*60 + "\n")
            self._write("SUMMARY:\n\n")
            self._write(" Totals: %s\t%0.4fs\n\n" % (len(self._failed_testresults) + len(self._passed_testresults),
                                                     (end_time-self._start_time).total_seconds()))
            
            self._write(" Passed: %s\n" % len(self._passed_testresults))
            for it in self._passed_testresults:
                self._write(" \t%s\t%0.4fs\n" % (it.testcase.test_name,
                                                 it.end_time-it.begin_time))
            self._write("\n")
            
            self._write(" Failed: %s\n" % len(self._failed_testresults))
            for it in self._failed_testresults:
                self._write_err(" \t%s\t%0.4fs\n" % (it.testcase.test_name,
                                                     it.end_time-it.begin_time))
            
    def log_test_result(self, testcase, testresult ):
        '''记录一个测试结果
        :param testcase: 测试用例
        :type testcase: TestCase
        :param testresult: 测试结果
        :type testresult: TestResult
        '''
        if testresult.passed:
            self._passed_testresults.append(testresult)
        else:
            self._failed_testresults.append(testresult)
        self._write("run test case: %s(pass?:%s)\n" % (testcase.test_name, testresult.passed))
                    
    def log_record(self, level, tag, msg, record={}):
        '''增加一个记录
        :param level: 日志级别
        :param msg: 日志消息
        :param tag: 日志标签
        :param record: 日志记录信息
        :type level: string
        :type tag: string
        :type msg: string
        :type record: dict
        '''
        self._write("%s\n" % (msg))
    
    def log_filtered_test(self, loader, testcase, reason):
        '''记录一个被过滤的测试用例
        :param loader: 用例加载器
        :type loader: TestLoader
        :param testcase: 测试用例
        :type testcase: TestCase
        :param reason: 过滤原因
        :type reason: str
        '''
        self._write("filtered test case: %s (reason: %s)\n" % (testcase.test_name, reason))

    def log_load_error(self, loader, name, error):
        '''记录一个加载失败的用例或用例集
        :param loader: 用例加载器
        :type loader: TestLoader
        :param name: 名称
        :type name: str
        :param error: 错误信息
        :type error: str
        '''
        line = ""
        for line in reversed(error.split("\n")):
            if line.strip():
                break
        self._write_err("load test failed: %s (error: %s)\n" % (name, line))

    def get_testresult_factory(self):
        '''获取对应的TestResult工厂
        :returns ITestResultFactory
        '''
        if self._output_testresult:
            return StreamTestResultFactory(self._stream)
        else:
            return EmptyTestResultFactory()
        
    @classmethod
    def get_parser(cls):
        '''获取命令行参数解析器（如果实现）

        :returns: 解析器对象
        :rtype: argparse.ArgumentParser
        '''
        parser = argparse.ArgumentParser(usage=report_usage)
        parser.add_argument("--no-output-result", action="store_true", help="don't output detail result of test cases")
        parser.add_argument("--no-summary", action="store_true", help="don't output summary information")
        return parser

    @classmethod
    def parse_args(cls, args_string):
        '''通过命令行参数构造对象
        
        :returns: 测试报告
        :rtype: cls
        '''
        args = cls.get_parser().parse_args(args_string)
        return cls(
            output_testresult=not args.no_output_result,
            output_summary=not args.no_summary)

class XMLTestResultFactory(ITestResultFactory):
    '''XML形式TestResult工厂
    '''
    def create(self, testcase):
        '''创建TestResult对象
        :param testcase: 测试用例
        :type testcase: TestCase
        :return TestResult
        '''
        return testresult.XmlResult(testcase)

class XMLTestReport(ITestReport):
    '''XML形式的测试报告
    '''
    def __init__(self):
        '''构造函数
        '''
        self._xmldoc = dom.Document()
        self._xmldoc.appendChild(self._xmldoc.createProcessingInstruction("xml-stylesheet", 'type="text/xsl" href="TestReport.xsl"'))
        self._runrstnode = self._xmldoc.createElement("RunResult")
        self._xmldoc.appendChild(self._runrstnode)
        self._result_factory = XMLTestResultFactory()
        
    def begin_report(self):
        '''开始测试执行
        '''
        self._time_start = datetime.now()
        
        xmltpl = "<TestEnv><PC>%s</PC><OS>%s</OS></TestEnv>"
        hostname = socket.gethostname()
        os_ver = get_os_version()
        envxml = dom.parseString(xmltpl % (hostname, os_ver))
        self._runrstnode.appendChild(envxml.childNodes[0])
        
    def end_report(self):
        '''结束测试执行
        :param passed: 测试是否通过
        :type passed: boolean
        '''
        time_end = datetime.now()
        timexml = "<RunTime><StartTime>%s</StartTime><EndTime>%s</EndTime><Duration>%s</Duration></RunTime>"
        timexml = timexml % (self._time_start.strftime("%Y-%m-%d %H:%M:%S"), time_end.strftime("%Y-%m-%d %H:%M:%S"), str(time_end-self._time_start).split('.')[0] )
        timenodes = dom.parseString(timexml)
        self._runrstnode.appendChild(timenodes.childNodes[0])
        
        xmldata = to_pretty_xml(self._xmldoc)
        with codecs_open('TestReport.xml', 'wb') as fd:
            fd.write(xmldata)
        test_repor_xsl = get_inner_resource("statics", "TestReport.xsl")
        shutil.copy(test_repor_xsl, os.getcwd())
        test_result_xsl = get_inner_resource("statics", "TestResult.xsl")
        shutil.copy(test_result_xsl, os.getcwd())
    
    def log_test_result(self, testcase, testresult ):
        '''记录一个测试结果
        :param testcase: 测试用例
        :type testcase: TestCase
        :param testresult: 测试结果
        :type testresult: XmlResult
        '''
        casemark = saxutils.escape(testcase.test_doc)
        nodestr = """<TestResult result="%s" log="%s" status="%s">%s</TestResult>
        """ % (testresult.passed, testresult.file_path, testcase.status, casemark)
        doc2 = dom.parseString(smart_binary(nodestr))
        resultNode = doc2.childNodes[0]
        resultNode.setAttribute("name", smart_text(saxutils.escape(testcase.test_name)))
        resultNode.setAttribute("owner", smart_text(saxutils.escape(testcase.owner)))
        self._runrstnode.appendChild(resultNode)
    
    def log_record(self, level, tag, msg, record={}):
        '''增加一个记录
        :param level: 日志级别
        :param msg: 日志消息
        :param tag: 日志标签
        :param record: 日志记录信息
        :type level: string
        :type tag: string
        :type msg: string
        :type record: dict
        '''        
        if tag == 'LOADER' and level == EnumLogLevel.ERROR:
            if 'error_testname' in record and 'error' in record:
                testname = record['error_testname']
                mdfailsnode = self._xmldoc.createElement("LoadFailure")
                self._runrstnode.appendChild(mdfailsnode)
                logfile = '%s.log' % testname
                xmltpl = """<Module name="%s" log="%s"/>""" % (testname, logfile)
                mdfailsnode.appendChild(dom.parseString(xmltpl).childNodes[0])
                with codecs_open(logfile, 'w', encoding="utf-8") as fd:
                    fd.write(record['error'])
                
    def log_filtered_test(self, loader, testcase, reason):
        '''记录一个被过滤的测试用例
        :param loader: 用例加载器
        :type loader: TestLoader
        :param testcase: 测试用例
        :type testcase: TestCase
        :param reason: 过滤原因
        :type reason: str
        '''
        nodestr = """<FilterTest name="%s" reason="%s"></FilterTest>
        """ % (
            smart_text(saxutils.escape(testcase.test_name)),
            smart_text(saxutils.escape(reason))
        )
        doc2 = dom.parseString(nodestr)
        filterNode = doc2.childNodes[0]
        self._runrstnode.appendChild(filterNode)

    def log_load_error(self, loader, name, error):
        '''记录一个加载失败的用例或用例集
        :param loader: 用例加载器
        :type loader: TestLoader
        :param name: 名称
        :type name: str
        :param error: 错误信息
        :type error: str
        '''
        log_file = "%s.log" % name
        nodestr = """<LoadTestError name="%s" log="%s"></LoadTestError>
        """ % (
            smart_text(saxutils.escape(name)),
            log_file,
        )
        doc2 = dom.parseString(nodestr)
        errNode = doc2.childNodes[0]
        self._runrstnode.appendChild(errNode)
        with codecs_open(log_file, 'w', encoding="utf-8") as fd:
            fd.write(error)

    def get_testresult_factory(self):
        '''获取对应的TestResult工厂
        :returns ITestResultFactory
        '''
        return self._result_factory
    
    @classmethod
    def get_parser(cls):
        '''获取命令行参数解析器（如果实现）

        :returns: 解析器对象
        :rtype: argparse.ArgumentParser
        '''
        return argparse.ArgumentParser(usage=report_usage)

    @classmethod
    def parse_args(cls, args_string):
        '''通过命令行参数构造对象
        
        :returns: 测试报告
        :rtype: cls
        '''
        return cls()

class JSONTestResultFactory(ITestResultFactory):
    '''JSON形式TestResult工厂
    '''
                    
    def create(self, testcase):
        '''创建TestResult对象
        :param testcase: 测试用例
        :type testcase: TestCase
        :return TestResult
        '''
        return testresult.JSONResult(testcase)
    
class JSONTestReportBase(ITestReport):
    '''JSON格式的测试报告基类
    '''
    def __init__(self, title="调试测试"):
        '''构造函数

        :param title: 报告标题
        :type title: str
        '''
        self._results = []
        self._logs = []
        self._filtered_tests = []
        self._load_errors = []
        self._testcases = []
        self._data = {
            "version": "1.0",
            "summary": {
                "tool": "QTA",
                "title": title,
            },
            "results": self._results,
            "logs": self._logs,
            "filtered_tests": self._filtered_tests,
            "load_errors": self._load_errors,
            "loaded_testcases": self._testcases,
        }
        self._testcase_names = set()
        self._testcase_total_run = 0
        self._testcase_passed = 0
        self._testcase_total_count = 0
        
    def begin_report(self):
        '''开始测试执行
        '''
        self._data["summary"]["start_time"] = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        self._data["summary"]["hostname"] = socket.gethostname()
        self._data["summary"]["os"] = get_os_version()

    def end_report(self):
        '''结束测试执行
        :param passed: 测试是否通过
        :type passed: boolean
        '''
        self._data["summary"]["testcase_total_run"] = self._testcase_total_run
        self._data["summary"]["testcase_total_count"] = self._testcase_total_count
        self._data["summary"]["testcase_passed"] = self._testcase_passed
        self._data["summary"]["succeed"] = self._testcase_passed == self._testcase_total_count
        self._data["summary"]["end_time"] = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        
    def log_test_result(self, testcase, testresult ):
        '''记录一个测试结果
        :param testcase: 测试用例
        :type testcase: TestCase
        :param testresult: 测试结果
        :type testresult: TestResult
        '''
        self._testcase_total_run += 1
        if testcase.test_name not in self._testcase_names:
            self._testcase_names.add(testcase.test_name)
            self._testcase_total_count += 1
            
        if testresult.passed:
            self._testcase_passed += 1
        self._results.append(testresult.get_file())
    
    def log_record(self, level, tag, msg, record):
        '''增加一个记录
        :param level: 日志级别
        :param msg: 日志消息
        :param tag: 日志标签
        :param record: 日志记录信息
        :type level: string
        :type tag: string
        :type msg: string
        :type record: dict
        '''
        self._logs.append({
            "level": level,
            "tag": tag,
            "message": msg,
            "record": record
        })

    def log_loaded_tests(self, loader, testcases):
        '''记录加载成功的用例

        :param loader: 用例加载器
        :type loader: TestLoader
        :param testcases: 测试用例列表
        :type testcases: list
        '''
        self._testcases += [
            {"name": testcase.test_name} 
            for testcase in testcases
        ]

    def log_filtered_test(self, loader, testcase, reason):
        '''记录一个被过滤的测试用例
        :param loader: 用例加载器
        :type loader: TestLoader
        :param testcase: 测试用例
        :type testcase: TestCase
        :param reason: 过滤原因
        :type reason: str
        '''
        self._filtered_tests.append({
            "name": testcase.test_name,
            "reason": reason
        })

    def log_load_error(self, loader, name, error):
        '''记录一个加载失败的用例或用例集
        :param loader: 用例加载器
        :type loader: TestLoader
        :param name: 名称
        :type name: str
        :param error: 错误信息
        :type error: str
        '''
        self._load_errors.append({
            "name": name,
            "error": error
        })

class JSONTestReport(JSONTestReportBase):
    '''JSON格式的测试报告
    '''
    def __init__(self, fd, title="调试测试"):
        '''构造函数

        :param fd: 输出流
        :type fd: file object
        :param title: 报告标题
        :type title: str
        '''
        self._fd = fd
        super(JSONTestReport, self).__init__(title)

    def end_report(self):
        '''结束测试执行
        '''
        super(JSONTestReport, self).end_report()
        json.dump(self._data, self._fd)

    def get_testresult_factory(self):
        '''获取对应的TestResult工厂
        :returns ITestResultFactory
        '''
        return JSONTestResultFactory()

    @classmethod
    def get_parser(cls):
        '''获取命令行参数解析器（如果实现）

        :returns: 解析器对象
        :rtype: argparse.ArgumentParser
        '''
        parser = argparse.ArgumentParser(usage=report_usage)
        parser.add_argument("--title", help="report title", default="Debug test report")
        parser.add_argument("-o", "--output", help="output file name", required=True)
        return parser

    @classmethod
    def parse_args(cls, args_string):
        '''通过命令行参数构造对象
        
        :returns: 测试报告
        :rtype: cls
        '''
        args = cls.get_parser().parse_args(args_string)
        fd = codecs_open(args.output, 'w', encoding="utf-8")
        return cls(fd, title=args.title)
    
class HtmlTestReport(JSONTestReportBase):
    """html test report
    """
    def get_testresult_factory(self):
        '''获取对应的TestResult工厂
        :returns ITestResultFactory
        '''
        return HtmlTestResultFactory()
    
    def end_report(self):
        super(HtmlTestReport, self).end_report()
        data = json.dumps(self._data)
        content = "let qta_report_data = %s" % data
        with codecs_open("qta-report.js", "w", encoding="utf-8") as fd:
            fd.write(content)
        
        qta_report_html = get_inner_resource("statics", "qta-report.html")
        shutil.copy(qta_report_html, os.getcwd())
        
    @classmethod
    def get_parser(cls):
        '''获取命令行参数解析器（如果实现）

        :returns: 解析器对象
        :rtype: argparse.ArgumentParser
        '''
        parser = argparse.ArgumentParser(usage=report_usage)
        parser.add_argument("--title", help="report title", default="Debug test report")
        return parser

    @classmethod
    def parse_args(cls, args_string):
        '''通过命令行参数构造对象
        
        :returns: 测试报告
        :rtype: cls
        '''
        args = cls.get_parser().parse_args(args_string)
        return cls(title=args.title)
        
class HtmlTestResultFactory(ITestResultFactory):
    """html test result factory
    """
    def create(self, testcase):
        return testresult.HtmlResult(testcase)


def __init_report_types():
    global report_types
    if report_types:
        return
    report_types.update({
        "empty"  : EmptyTestReport,
        "stream" : StreamTestReport,
        "xml"    : XMLTestReport,
        "json"   : JSONTestReport,
        "html"   : HtmlTestReport,
    })

    # Register other `ITestReport` implementiations from entry points 
    for ep in pkg_resources.iter_entry_points(REPORT_ENTRY_POINT):
        if ep.name not in report_types:
            try:
                report_types[ep.name] = ep.load()
            except:
                stack = traceback.format_exc()
                logger.warn("load ITestReport entry point for %s failed:\n" % (ep.name, stack))

__init_report_types()
del __init_report_types



# -*- coding: utf-8 -*-
#
# Tencent is pleased to support the open source community by making QTA available.
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the BSD 3-Clause License (the "License"); you may not use this 
# file except in compliance with the License. You may obtain a copy of the License at
# 
# https://opensource.org/licenses/BSD-3-Clause
# 
# Unless required by applicable law or agreed to in writing, software distributed 
# under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#
'''
测试用例序列化和反序列化
'''

import sys
import pickle
from testbase.testcase import TestSuite


class _EmptyClass(object):
    pass

def dumps( testcase ):
    """序列化测试用例
    
    :param testcase: 测试用例
    :type testcase: TestCase
    """
    if isinstance(testcase, TestSuite):
        return {
            'id': testcase.suite_class_name,
            'data': pickle.dumps(testcase.dumps())
        }
    else:
        return {
            'id': testcase.test_class_name,
            'data': pickle.dumps(testcase.casedata),
            'dname': testcase.casedataname,
        }

def loads( buf ):
    """反序列化测试用例
    
    :param buf: 测试用例序列化数据
    :type buf: dict
    :returns: TestCase
    """
    testname = buf['id']
    items = testname.split('.')
    classname = items[-1]
    modulename = '.'.join(items[0:-1])
    __import__(modulename)
    module = sys.modules[modulename]
    testclass = getattr(module, classname)
    data = pickle.loads(buf['data'])
    if issubclass(testclass, TestSuite):
        obj = _EmptyClass()
        obj.__class__ = testclass
        obj.loads(data)
        return obj
    else:
        return testclass(data, buf['dname'])



# -*- coding: utf-8 -*-
#
# Tencent is pleased to support the open source community by making QTA available.
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the BSD 3-Clause License (the "License"); you may not use this 
# file except in compliance with the License. You may obtain a copy of the License at
# 
# https://opensource.org/licenses/BSD-3-Clause
# 
# Unless required by applicable law or agreed to in writing, software distributed 
# under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#
import six
'''
测试用例加载
'''

import types
import sys
import pkgutil
import os
import imp
import traceback

from testbase.testcase import TestCase, SeqTestSuite
from testbase.conf import settings
from testbase.util import smart_text
from testbase import datadrive

class TestLoader(object):
    '''测试用例加载器
    '''
    def __init__(self, filter_func=None ):
        '''构造函数
        :param filter: 用例过滤函数，函数原型为 filter_func(testcase_obj)，返回None/False表示不过滤此用例
        :type filter: callable
        '''
        self._filter = filter_func
        self._module_errs = {}
        self._filtered_tests = {}
            
    def get_last_errors(self):
        '''返回最后一次load调用时加载失败的全部模块和对应错误信息
        
        :returns dict: 模块和对应的错误信息
        '''
        return self._module_errs
                    
    def get_filtered_tests(self):
        '''返回最后一次load调用时被过滤掉的测试用例
        '''
        return list(self._filtered_tests.keys())
    
    def get_filtered_tests_with_reason(self):
        '''返回最后一次load调用时被过滤掉的测试用例和过滤原因
        '''
        return self._filtered_tests
                    
    def load(self, testname ):
        '''通过名字加载测试用例
        
        :param name: 用例或用例名称
        :type name: string/list
        :returns list - 测试用例对象列表
        '''

        if isinstance(testname, list):
            testnames = testname
        else:
            testnames = [testname]

        self._module_errs = {}
        testcases = []

        for testname in testnames:
            testname = smart_text(testname)
            if settings.DATA_DRIVE:
                self._dataset = TestDataLoader().load()
            if '/' in testname:
                testname, datakeyname = testname.split('/', 1)
            else:
                datakeyname = None 
                
            obj = self._load(testname)
            
            if isinstance(obj, types.ModuleType):
                if hasattr(obj, '__path__'):
                    testcases += self._load_from_package(obj)
                else:
                    testcases += self._load_from_module(obj)
            elif isinstance(obj, type):
                testcases += self._load_from_class(obj)
        
        #过滤掉重复的用例
        testcase_dict = {}
        for testcase in testcases:
            if datakeyname and smart_text(testcase.casedataname) != datakeyname:
                continue
            testcase_dict[testcase.test_name] = testcase
            
        return list(testcase_dict.values())

    def _load(self, testname ):
        '''加载对应的对象
        
        :param name: 用例或用例集名称
        :type name: string
        :returns - Type/ModuleType
        '''
        parts = testname.split('.')
        module = None
        parts_imp = parts[:]
        while parts_imp:
            try:
                modulename = '.'.join(parts_imp)
                __import__(modulename)   #__import__得到的是最外层模块的object
                module = sys.modules[modulename]
                break
            except ImportError:
                del parts_imp[-1]
            except:
                del parts_imp[-1]
                break
            
        obj = module
        
        if parts_imp == parts: #为一个包或模块
            return obj
        
        elif parts_imp == parts[0:-1] and hasattr(module, parts[-1]): #为一个类
            try:
                testclass = getattr(module, parts[-1])
                if not self._is_testcase_class(testclass):
                    raise TypeError("%s不是一个有效的测试用例"%testname)      
            except:
                self._module_errs[testname] = traceback.format_exc()
                return
            else:
                return testclass


        else: #触发异常
            if parts_imp:
                modulename = '.'.join(parts_imp)
                modulename += '.'
            else:
                modulename = ''
            modulename += parts[len(parts_imp)]
            try:
                __import__(modulename)
            except:
                self._module_errs[modulename] = traceback.format_exc()
    
    def _is_testcase_class(self, obj ):
        '''是否为测试用例类
        
        :returns bool - 是否为用例类
        '''
        return ( isinstance(obj, type) and
              issubclass(obj, TestCase) and
              hasattr(obj, "runTest") and 
              getattr(obj, "priority", None))
        
    def _walk_package_error(self, modulename ):
        '''walk_packages错误回调
        '''
        self._module_errs[modulename] = traceback.format_exc()
        
    def _load_from_package(self, pkg ):
        '''从一个python包加载测试用例
        
        :param pkg: Python包
        :type pkg: ModuleType
        :returns list - 测试用例对象列表
        '''
        tests = []
        for _, modulename, ispkg in pkgutil.walk_packages(pkg.__path__, pkg.__name__+'.', onerror=self._walk_package_error):
            if ispkg:
                continue
            try:
                __import__(modulename)
                tests += self._load_from_module(sys.modules[modulename])
            except:
                self._module_errs[modulename] = traceback.format_exc()
        return tests
    
    def _load_from_module(self, mod):
        '''从一个python模块加载测试用例
        
        :param mod: Python模块
        :type mod: ModuleType
        :returns list - 测试用例对象列表
        '''
        tests = []
        for name in dir(mod):
            obj = getattr(mod, name)
            if self._is_testcase_class(obj):
                tests += self._load_from_class(obj)
        if hasattr(mod, '__qtaf_seq_tests__'): #测试用例需要顺序执行
            seqdef = mod.__qtaf_seq_tests__
            if not isinstance(seqdef, list):
                raise TypeError("__qtaf_seq_tests__必须为list类型")
            if len(seqdef) == 0:
                raise ValueError("__qtaf_seq_tests__必须至少包含一个测试用例")
#            modulename = mod.__name__
            for it in seqdef:
                if type(it) != type(TestCase) or not issubclass(it, TestCase):
                    raise TypeError("__qtaf_seq_tests__的元素必须为测试用例类")
#                 if it.__module__ != modulename:
#                     raise ValueError("__qtaf_seq_tests__中的测试用例类必须在当前模块中定义，'%s'不属于当前模块" % it.__name__)
            test_dict = {}
            for test in tests:
                test_dict[type(test)] = test
            for it in seqdef:
                if it not in test_dict:
                    raise RuntimeError("__qtaf_seq_tests__中的测试用例'%s'已被过滤"%it.__name__)
            tests = [SeqTestSuite([ test_dict[it] for it in seqdef])]
        return tests
    
    def _load_from_class(self, cls ):
        '''加载用例类
        
        :param cls: Python类 
        :type cls: Type
        :returns list - 测试用例对象列表
        '''
        tests = []
        if datadrive.is_datadrive(cls):
            tests = datadrive.load_datadrive_tests(cls)
        elif settings.DATA_DRIVE:
            tests = [cls(self._dataset[it], str(it)) for it in self._dataset]
        else:
            tests = [cls()]
        
        if self._filter:
            final_tests = []
            for it in tests:
                filter_reason = self._filter(it)
                if filter_reason:
                    self._filtered_tests[it] = filter_reason
                else:
                    final_tests.append(it)
            tests = final_tests 
        return tests
    
class TestDataLoader(object):
    '''测试数据加载器
    '''
    MODULE_NAME = 'testbase.loader.testdata'
    
    
    def _load_dataset_module_from_file(self):
        '''从文件中加载数据模块
        '''
        if self.MODULE_NAME in sys.modules:
            return sys.modules[self.MODULE_NAME]
        
        if os.path.isfile(__file__):
            qtaf_top_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), '..')
            top_dir = qtaf_top_dir
        else: #使用的egg包
            qtaf_top_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), '..')
            top_dir = os.path.join(qtaf_top_dir, '..', '..')
        top_dir = os.path.abspath(top_dir)
        
        py_path = os.path.join(top_dir, settings.DATA_SOURCE)
        if not os.path.isfile(py_path):
            raise RuntimeError("指定的数据源文件\"%s\"不存在"%py_path)
            
        module_name = os.path.basename(py_path)
        module_name = module_name[0:module_name.rfind('.')]
        
        fd, pathname, desc = imp.find_module(module_name, [os.path.dirname(py_path), top_dir])
        module = imp.load_module(self.MODULE_NAME, fd, pathname, desc)
        return module
             
    def _load_dataset_module_from_name(self):
        '''根据模块名加载模块
        '''
        __import__(settings.DATA_SOURCE)
        return sys.modules[settings.DATA_SOURCE]
        
    def _load_dataset(self):
        '''加载数据
        '''
        if not settings.DATA_SOURCE:
            raise RuntimeError("DATA_DRIVE=True，但未指定的数据源文件")
        
        if isinstance(settings.DATA_SOURCE, six.string_types):
            if settings.DATA_SOURCE.endswith('.py'):
                module = self._load_dataset_module_from_file()
            else:
                module = self._load_dataset_module_from_name()
            if not hasattr(module, 'DATASET'):
                raise RuntimeError("数据源文件\"%s\"没有定义模块变量\"DATASET\"" % module)         
            return module.DATASET
        
        else:
            return settings.DATA_SOURCE

    def load(self):
        '''从数据源加载测试数据
        :returns list - 测试数据集
        '''
        dataset = self._load_dataset()
        if isinstance(dataset, list) or isinstance(dataset, tuple):
            dataset_dict = {}
            for idx, it in enumerate(dataset):
                dataset_dict[str(idx)] = it
            dataset = dataset_dict
        elif isinstance(dataset, dict):
            pass
        else:
            raise ValueError("DATASET必须为list/tuple/dict类型")
        if len(dataset) == 0:
            raise ValueError("DATASET不可以为空")
        return dataset
    


# -*- coding: utf-8 -*-
#
# Tencent is pleased to support the open source community by making QTA available.
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the BSD 3-Clause License (the "License"); you may not use this 
# file except in compliance with the License. You may obtain a copy of the License at
# 
# https://opensource.org/licenses/BSD-3-Clause
# 
# Unless required by applicable law or agreed to in writing, software distributed 
# under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#
"""a module for unittest
"""
import os

from testbase.conf import settings

class _NotExistedItem(object):
    pass

class modify_settings(object):
    """temporarily modify settings
    """
    def __init__(self, **kwargs):
        self.new_conf = kwargs
        self.old_conf = {}
        
    def __enter__(self):
        settings._Settings__sealed = False
        for key in self.new_conf:
            self.old_conf[key] = getattr(settings, key, _NotExistedItem())
            if isinstance(self.old_conf[key], _NotExistedItem):
                settings._Settings__keys.add(key)
            setattr(settings, key, self.new_conf[key])
        
    def __exit__(self, *args):
        for key in self.old_conf:
            old_value = self.old_conf[key]
            if isinstance(old_value, _NotExistedItem):
                delattr(settings, key)
                settings._Settings__keys.remove(key)
            else:
                setattr(settings, key, old_value)
        settings._Settings__sealed = True
        
class modify_environ(object):
    """temporarily modify envrion
    """
    def __init__(self, **kwargs):
        self.new_conf = kwargs
        self.old_conf = {}
        
    def __enter__(self):
        for key in self.new_conf:
            self.old_conf[key] = os.environ.get(key, _NotExistedItem())
            os.environ[key] = self.new_conf[key]
        
    def __exit__(self, *args):
        for key in self.old_conf:
            old_value = self.old_conf[key]
            if isinstance(old_value, _NotExistedItem):
                del os.environ[key]
            else:
                os.environ[key] = old_value
